@@ -53,7 +53,7 @@
 #define CUBE_IN_DEFAULT FALSE
 
 static char *cubeImages[] = {
-    "freedesktop.png"
+    "freedesktop"
 };
 #define N_CUBE_IMAGES (sizeof (cubeImages) / sizeof (cubeImages[0]))
 
@@ -130,8 +130,9 @@ typedef struct _CubeScreen {
     DonePaintScreenProc	       donePaintScreen;
     PaintScreenProc	       paintScreen;
     PaintTransformedScreenProc paintTransformedScreen;
-    PaintBackgroundProc	       paintBackground;
+    ApplyScreenTransformProc   applyScreenTransform;
     SetScreenOptionProc	       setScreenOption;
+    OutputChangeNotifyProc     outputChangeNotify;
 
     CompOption opt[CUBE_SCREEN_OPTION_NUM];
 
@@ -165,6 +166,19 @@ typedef struct _CubeScreen {
     int		    imgCurFile;
     CompOptionValue *imgFiles;
 
+    int nOutput;
+    int output[64];
+    int outputMask[64];
+
+    Bool cleared[64];
+
+    Bool fullscreenOutput;
+
+    float outputXScale;
+    float outputYScale;
+    float outputXOffset;
+    float outputYOffset;
+
 #ifdef USE_LIBRSVG
     cairo_t	    *cr;
 #endif
@@ -232,8 +246,7 @@ readSvgToTexture (CompScreen   *s,
 
     CUBE_SCREEN (s);
 
-    if (!openImageFile (svgFileName, &name, NULL))
-	return FALSE;
+    name = strdup (svgFileName);
 
     svgHandle = rsvg_handle_new_from_file (name, &error);
 
@@ -249,8 +262,8 @@ readSvgToTexture (CompScreen   *s,
 
     if (cs->opt[CUBE_SCREEN_OPTION_SCALE_IMAGE].value.b)
     {
-	pw = s->width;
-	ph = s->height;
+	pw = (cs->nOutput > 1) ? s->outputDev[0].width  : s->width;
+	ph = (cs->nOutput > 1) ? s->outputDev[0].height : s->height;
     }
     else
     {
@@ -329,10 +342,22 @@ cubeLoadImg (CompScreen *s,
 	     int	n)
 {
     unsigned int width, height;
+    int		 pw, ph;
 
     CUBE_SCREEN (s);
 
-    if (!cs->imgNFile || cs->pw != s->width || cs->ph != s->height)
+    if (!cs->fullscreenOutput)
+    {
+	pw = s->width;
+	ph = s->height;
+    }
+    else
+    {
+	pw = s->outputDev[0].width;
+	ph = s->outputDev[0].height;
+    }
+
+    if (!cs->imgNFile || cs->pw != pw || cs->ph != ph)
     {
 	finiTexture (s, &cs->texture);
 	initTexture (s, &cs->texture);
@@ -389,10 +414,10 @@ cubeLoadImg (CompScreen *s,
     }
     else
     {
-	float x1 = width  / 2.0f - s->width  / 2.0f;
-	float y1 = height / 2.0f - s->height / 2.0f;
-	float x2 = width  / 2.0f + s->width  / 2.0f;
-	float y2 = height / 2.0f + s->height / 2.0f;
+	float x1 = width  / 2.0f - pw / 2.0f;
+	float y1 = height / 2.0f - ph / 2.0f;
+	float x2 = width  / 2.0f + pw / 2.0f;
+	float y2 = height / 2.0f + ph / 2.0f;
 
 	cs->tc[2] = COMP_TEX_COORD_X (&cs->texture.matrix, x2);
 	cs->tc[3] = COMP_TEX_COORD_Y (&cs->texture.matrix, y1);
@@ -422,6 +447,8 @@ cubeUpdateGeometry (CompScreen *s,
 
     CUBE_SCREEN (s);
 
+    sides *= cs->nOutput;
+
     distance = 0.5f / tanf (M_PI / sides);
     radius   = 0.5f / sinf (M_PI / sides);
 
@@ -467,6 +494,94 @@ cubeUpdateGeometry (CompScreen *s,
     return TRUE;
 }
 
+static void
+cubeUpdateOutputs (CompScreen *s)
+{
+    BoxPtr pBox0, pBox1;
+    int    i, j, k, x;
+
+    CUBE_SCREEN (s);
+
+    k = 0;
+
+    cs->fullscreenOutput = TRUE;
+
+    for (i = 0; i < s->nOutputDev; i++)
+    {
+	cs->outputMask[i] = -1;
+
+	/* dimensions must match first output */
+	if (s->outputDev[i].width  != s->outputDev[0].width ||
+	    s->outputDev[i].height != s->outputDev[0].height)
+	    continue;
+
+	pBox0 = &s->outputDev[0].region.extents;
+	pBox1 = &s->outputDev[i].region.extents;
+
+	/* top and bottom line must match first output */
+	if (pBox0->y1 != pBox1->y1 || pBox0->y2 != pBox1->y2)
+	    continue;
+
+	k++;
+
+	for (j = 0; j < s->nOutputDev; j++)
+	{
+	    pBox0 = &s->outputDev[j].region.extents;
+
+	    /* must not intersect other output region */
+	    if (i != j && pBox0->x2 > pBox1->x1 && pBox0->x1 < pBox1->x2)
+	    {
+		k--;
+		break;
+	    }
+	}
+    }
+
+    if (k != s->nOutputDev)
+    {
+	cs->fullscreenOutput = FALSE;
+	cs->nOutput = 1;
+	return;
+    }
+
+    /* add output indices from left to right */
+    j = 0;
+    for (;;)
+    {
+	x = MAXSHORT;
+	k = -1;
+
+	for (i = 0; i < s->nOutputDev; i++)
+	{
+	    if (cs->outputMask[i] != -1)
+		continue;
+
+	    if (s->outputDev[i].region.extents.x1 < x)
+	    {
+		x = s->outputDev[i].region.extents.x1;
+		k = i;
+	    }
+	}
+
+	if (k < 0)
+	    break;
+
+	cs->outputMask[k] = j;
+	cs->output[j]     = k;
+
+	j++;
+    }
+
+    cs->nOutput = j;
+
+    if (cs->nOutput == 1)
+    {
+	if (s->outputDev[0].width  != s->width ||
+	    s->outputDev[0].height != s->height)
+	    cs->fullscreenOutput = FALSE;
+    }
+}
+
 static CompOption *
 cubeGetScreenOptions (CompScreen *screen,
 		      int	 *count)
@@ -1053,6 +1168,8 @@ cubePreparePaintScreen (CompScreen *s,
 	}
     }
 
+    memset (cs->cleared, 0, sizeof (Bool) * s->nOutputDev);
+
     UNWRAP (cs, s, preparePaintScreen);
     (*s->preparePaintScreen) (s, msSinceLastPaint);
     WRAP (cs, s, preparePaintScreen, cubePreparePaintScreen);
@@ -1062,6 +1179,7 @@ cubePreparePaintScreen (CompScreen *s,
 static Bool
 cubePaintScreen (CompScreen		 *s,
 		 const ScreenPaintAttrib *sAttrib,
+		 const CompTransform	*transform,
 		 Region			 region,
 		 int			 output,
 		 unsigned int		 mask)
@@ -1077,7 +1195,7 @@ cubePaintScreen (CompScreen		 *s,
     }
 
     UNWRAP (cs, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (cs, s, paintScreen, cubePaintScreen);
 
     return status;
@@ -1096,46 +1214,126 @@ cubeDonePaintScreen (CompScreen *s)
     WRAP (cs, s, donePaintScreen, cubeDonePaintScreen);
 }
 
+static void
+cubeMoveViewportAndPaint (CompScreen		  *s,
+			  const ScreenPaintAttrib *sAttrib,
+			  const CompTransform	  *transform,
+			  int			  output,
+			  unsigned int		  mask,
+			  int			  dx)
+{
+    CUBE_SCREEN (s);
+
+    if (cs->nOutput > 1)
+    {
+	int cubeOutput, dView;
+
+	/* translate to cube output */
+	cubeOutput = cs->outputMask[output];
+
+	/* convert from window movement to viewport movement */
+	dView = -dx;
+
+	cubeOutput += dView;
+
+	dView      = cubeOutput / cs->nOutput;
+	cubeOutput = cubeOutput % cs->nOutput;
+
+	if (cubeOutput < 0)
+	{
+	    cubeOutput += cs->nOutput;
+	    dView--;
+	}
+
+	/* translate back to compiz output */
+	output = cs->output[cubeOutput];
+
+	moveScreenViewport (s, -dView, 0, FALSE);
+	(*s->paintTransformedScreen) (s, sAttrib, transform,
+				      &s->outputDev[output].region,
+				      output, mask);
+	moveScreenViewport (s, dView, 0, FALSE);
+    }
+    else
+    {
+	moveScreenViewport (s, dx, 0, FALSE);
+	(*s->paintTransformedScreen) (s, sAttrib, transform, &s->region,
+				      output, mask);
+	moveScreenViewport (s, -dx, 0, FALSE);
+    }
+}
+
 static void
 cubePaintTransformedScreen (CompScreen		    *s,
 			    const ScreenPaintAttrib *sAttrib,
+			    const CompTransform	    *transform,
+			    Region		    region,
 			    int			    output,
 			    unsigned int	    mask)
 {
     ScreenPaintAttrib sa = *sAttrib;
-    int		      xMove = 0;
-    float	      size = s->hsize;
+    int		      hsize, xMove = 0;
+    float	      size;
+    Bool	      clear;
 
     CUBE_SCREEN (s);
 
-    if (cs->sky.name)
+    hsize = s->hsize * cs->nOutput;
+    size  = hsize;
+
+    if (!cs->fullscreenOutput)
     {
-	if (mask & PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK)
-	    glClear (GL_STENCIL_BUFFER_BIT);
+	cs->outputXScale = (float) s->width / s->outputDev[output].width;
+	cs->outputYScale = (float) s->height / s->outputDev[output].height;
 
-	screenLighting (s, FALSE);
+	cs->outputXOffset =
+	    (s->width / 2.0f -
+	     (s->outputDev[output].region.extents.x1 +
+	      s->outputDev[output].region.extents.x2) / 2.0f) /
+	    (float) s->outputDev[output].width;
 
-	glPushMatrix ();
-
-	if (cs->animateSkyDome && cs->grabIndex == 0)
-	{
-	    glRotatef (sAttrib->xRotate, 0.0f, 1.0f, 0.0f);
-	    glRotatef (sAttrib->vRotate / 5.0f + 90.0f, 1.0f, 0.0f, 0.0f);
-	}
-	else
-	{
-	    glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
-	}
-
-	glCallList (cs->skyListId);
-	glPopMatrix ();
+	cs->outputYOffset =
+	    (s->height / 2.0f -
+	     (s->outputDev[output].region.extents.y1 +
+	      s->outputDev[output].region.extents.y2) / 2.0f) /
+	    (float) s->outputDev[output].height;
     }
     else
     {
-	if (mask & PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK)
-	    glClear (GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
+	cs->outputXScale  = 1.0f;
+	cs->outputYScale  = 1.0f;
+	cs->outputXOffset = 0.0f;
+	cs->outputYOffset = 0.0f;
+    }
+
+    clear = cs->cleared[output];
+    if (!clear)
+    {
+	if (cs->sky.name)
+	{
+	    screenLighting (s, FALSE);
+
+	    glPushMatrix ();
+
+	    if (cs->animateSkyDome && cs->grabIndex == 0)
+	    {
+		glRotatef (sAttrib->xRotate, 0.0f, 1.0f, 0.0f);
+		glRotatef (sAttrib->vRotate / 5.0f + 90.0f, 1.0f, 0.0f, 0.0f);
+	    }
+	    else
+	    {
+		glRotatef (90.0f, 1.0f, 0.0f, 0.0f);
+	    }
+
+	    glCallList (cs->skyListId);
+	    glPopMatrix ();
+	}
 	else
-	    glClear (GL_COLOR_BUFFER_BIT);
+	{
+	    clearTargetOutput (s->display, GL_COLOR_BUFFER_BIT);
+	}
+
+	cs->cleared[output] = TRUE;
     }
 
     mask &= ~PAINT_SCREEN_CLEAR_MASK;
@@ -1162,18 +1360,18 @@ cubePaintTransformedScreen (CompScreen		    *s,
 	sa.xRotate = sAttrib->xRotate * cs->invert;
 	if (sa.xRotate > 0.0f)
 	{
-	    cs->xrotations = (int) (s->hsize * sa.xRotate) / 360;
-	    sa.xRotate = sa.xRotate - (360.0f * cs->xrotations) / s->hsize;
+	    cs->xrotations = (int) (hsize * sa.xRotate) / 360;
+	    sa.xRotate = sa.xRotate - (360.0f * cs->xrotations) / hsize;
 	}
 	else
 	{
-	    cs->xrotations = (int) (s->hsize * sa.xRotate) / 360;
+	    cs->xrotations = (int) (hsize * sa.xRotate) / 360;
 	    sa.xRotate = sa.xRotate -
-		(360.0f * cs->xrotations) / s->hsize + 360.0f / s->hsize;
+		(360.0f * cs->xrotations) / hsize + 360.0f / hsize;
 	    cs->xrotations--;
 	}
 
-	sa.xRotate = sa.xRotate / size * s->hsize;
+	sa.xRotate = sa.xRotate / size * hsize;
     }
     else
     {
@@ -1200,9 +1398,11 @@ cubePaintTransformedScreen (CompScreen		    *s,
 	}
     }
 
-    if (cs->grabIndex == 0 &&
+    if (!clear && cs->grabIndex == 0 && hsize > 2 &&
 	(cs->invert != 1 || sa.vRotate != 0.0f || sa.yTranslate != 0.0f))
     {
+	CompTransform sTransform = *transform;
+
 	screenLighting (s, TRUE);
 
 	glColor3usv (cs->color);
@@ -1212,17 +1412,25 @@ cubePaintTransformedScreen (CompScreen		    *s,
 	if (sAttrib->xRotate > 0.0f)
 	{
 	    sa.yRotate += 360.0f / size;
-	    translateRotateScreen (&sa);
+	    (*s->applyScreenTransform) (s, &sa, output, &sTransform);
+	    glLoadMatrixf (sTransform.m);
+	    glTranslatef (cs->outputXOffset, -cs->outputYOffset, 0.0f);
+	    glScalef (cs->outputXScale, cs->outputYScale, 1.0f);
 	    sa.yRotate -= 360.0f / size;
 	}
 	else
-	    translateRotateScreen (&sa);
+	{
+	    (*s->applyScreenTransform) (s, &sa, output, &sTransform);
+	    glLoadMatrixf (sTransform.m);
+	    glTranslatef (cs->outputXOffset, -cs->outputYOffset, 0.0f);
+	    glScalef (cs->outputXScale, cs->outputYScale, 1.0f);
+	}
 
 	glVertexPointer (3, GL_FLOAT, 0, cs->vertices);
 
 	glNormal3f (0.0f, -1.0f, 0.0f);
 
-	if (cs->invert == 1 && s->hsize == 4 && cs->texture.name)
+	if (cs->invert == 1 && hsize == 4 && cs->texture.name)
 	{
 	    enableTexture (s, &cs->texture, COMP_TEXTURE_FILTER_GOOD);
 	    glTexCoordPointer (2, GL_FLOAT, 0, cs->tc);
@@ -1252,23 +1460,22 @@ cubePaintTransformedScreen (CompScreen		    *s,
     /* outside cube */
     if (cs->invert == 1)
     {
-	if (cs->grabIndex || s->hsize > 4)
+	if (cs->grabIndex || hsize > 4)
 	{
 	    GLenum filter;
 	    int    i;
 
-	    xMove = cs->xrotations - ((s->hsize >> 1) - 1);
-	    sa.yRotate += (360.0f / size) * ((s->hsize >> 1) - 1);
+	    xMove = cs->xrotations - ((hsize >> 1) - 1);
+	    sa.yRotate += (360.0f / size) * ((hsize >> 1) - 1);
 
 	    filter = s->display->textureFilter;
 	    if (cs->grabIndex && cs->opt[CUBE_SCREEN_OPTION_MIPMAP].value.b)
 		s->display->textureFilter = GL_LINEAR_MIPMAP_LINEAR;
 
-	    for (i = 0; i < s->hsize; i++)
+	    for (i = 0; i < hsize; i++)
 	    {
-		moveScreenViewport (s, xMove, 0, FALSE);
-		(*s->paintTransformedScreen) (s, &sa, output, mask);
-		moveScreenViewport (s, -xMove, 0, FALSE);
+		cubeMoveViewportAndPaint (s, &sa, transform, output, mask,
+					  xMove);
 
 		sa.yRotate -= 360.0f / size;
 		xMove++;
@@ -1282,19 +1489,15 @@ cubePaintTransformedScreen (CompScreen		    *s,
 	    {
 		xMove = cs->xrotations;
 
-		moveScreenViewport (s, xMove, 0, FALSE);
-		(*s->paintTransformedScreen) (s, &sa, output, mask);
-		moveScreenViewport (s, -xMove, 0, FALSE);
+		cubeMoveViewportAndPaint (s, &sa, transform, output, mask,
+					  xMove);
 
 		xMove++;
-
-		moveScreenViewport (s, xMove, 0, FALSE);
 	    }
 
 	    sa.yRotate -= 360.0f / size;
 
-	    (*s->paintTransformedScreen) (s, &sa, output, mask);
-	    moveScreenViewport (s, -xMove, 0, FALSE);
+	    cubeMoveViewportAndPaint (s, &sa, transform, output, mask, xMove);
 	}
     }
     else
@@ -1319,20 +1522,19 @@ cubePaintTransformedScreen (CompScreen		    *s,
 
 	    if (sa.xRotate > 180.0f / size)
 	    {
-		xMove -= ((s->hsize >> 1) - 2);
-		sa.yRotate -= (360.0f / size) * ((s->hsize >> 1) - 2);
+		xMove -= ((hsize >> 1) - 2);
+		sa.yRotate -= (360.0f / size) * ((hsize >> 1) - 2);
 	    }
 	    else
 	    {
-		xMove -= ((s->hsize >> 1) - 1);
-		sa.yRotate -= (360.0f / size) * ((s->hsize >> 1) - 1);
+		xMove -= ((hsize >> 1) - 1);
+		sa.yRotate -= (360.0f / size) * ((hsize >> 1) - 1);
 	    }
 
-	    for (i = 0; i < s->hsize; i++)
+	    for (i = 0; i < hsize; i++)
 	    {
-		moveScreenViewport (s, xMove, 0, FALSE);
-		(*s->paintTransformedScreen) (s, &sa, output, mask);
-		moveScreenViewport (s, -xMove, 0, FALSE);
+		cubeMoveViewportAndPaint (s, &sa, transform, output, mask,
+					  xMove);
 
 		sa.yRotate += 360.0f / size;
 		xMove++;
@@ -1342,23 +1544,17 @@ cubePaintTransformedScreen (CompScreen		    *s,
 	}
 	else
 	{
-	    moveScreenViewport (s, xMove, 0, FALSE);
-	    (*s->paintTransformedScreen) (s, &sa, output, mask);
-	    moveScreenViewport (s, -xMove, 0, FALSE);
+	    cubeMoveViewportAndPaint (s, &sa, transform, output, mask, xMove);
 
 	    sa.yRotate += 360.0f / size;
 	    xMove = -cs->xrotations;
 
-	    moveScreenViewport (s, xMove, 0, FALSE);
-	    (*s->paintTransformedScreen) (s, &sa, output, mask);
-	    moveScreenViewport (s, -xMove, 0, FALSE);
+	    cubeMoveViewportAndPaint (s, &sa, transform, output, mask, xMove);
 
 	    sa.yRotate += 360.0f / size;
 	    xMove = 1 - cs->xrotations;
 
-	    moveScreenViewport (s, xMove, 0, FALSE);
-	    (*s->paintTransformedScreen) (s, &sa, output, mask);
-	    moveScreenViewport (s, -xMove, 0, FALSE);
+	    cubeMoveViewportAndPaint (s, &sa, transform, output, mask, xMove);
 	}
     }
 
@@ -1366,17 +1562,23 @@ cubePaintTransformedScreen (CompScreen		    *s,
 }
 
 static void
-cubePaintBackground (CompScreen   *s,
-		     Region	  region,
-		     unsigned int mask)
+cubeApplyScreenTransform (CompScreen		  *s,
+			  const ScreenPaintAttrib *sAttrib,
+			  int			  output,
+			  CompTransform	          *transform)
 {
     CUBE_SCREEN (s);
 
-    s->stencilRef++;
+    matrixTranslate (transform, cs->outputXOffset, -cs->outputYOffset, 0.0f);
+    matrixScale (transform, cs->outputXScale, cs->outputYScale, 1.0f);
 
-    UNWRAP (cs, s, paintBackground);
-    (*s->paintBackground) (s, region, mask);
-    WRAP (cs, s, paintBackground, cubePaintBackground);
+    UNWRAP (cs, s, applyScreenTransform);
+    (*s->applyScreenTransform) (s, sAttrib, output, transform);
+    WRAP (cs, s, applyScreenTransform, cubeApplyScreenTransform);
+
+    matrixScale (transform, 1.0f / cs->outputXScale,
+		 1.0f / cs->outputYScale, 1.0f);
+    matrixTranslate (transform, -cs->outputXOffset, cs->outputYOffset, 0.0f);
 }
 
 static Bool
@@ -1396,6 +1598,9 @@ cubeUnfold (CompDisplay     *d,
     {
 	CUBE_SCREEN (s);
 
+	if (s->hsize * cs->nOutput < 4)
+	    return FALSE;
+
 	if (otherScreenGrabExist (s, "rotate", "switcher", "cube", 0))
 	    return FALSE;
 
@@ -1503,6 +1708,22 @@ cubePrevImage (CompDisplay     *d,
     return FALSE;
 }
 
+static void
+cubeOutputChangeNotify (CompScreen *s)
+{
+    CUBE_SCREEN (s);
+
+    cubeUpdateOutputs (s);
+    cubeUpdateGeometry (s, s->hsize, cs->invert);
+
+    if (cs->imgNFile)
+	cubeLoadImg (s, cs->imgCurFile);
+
+    UNWRAP (cs, s, outputChangeNotify);
+    (*s->outputChangeNotify) (s);
+    WRAP (cs, s, outputChangeNotify, cubeOutputChangeNotify);
+}
+
 static Bool
 cubeSetGlobalScreenOption (CompScreen      *s,
 			   char		   *name,
@@ -1516,7 +1737,7 @@ cubeSetGlobalScreenOption (CompScreen      *s,
     status = (*s->setScreenOption) (s, name, value);
     WRAP (cs, s, setScreenOption, cubeSetGlobalScreenOption);
 
-    if (status && strcmp (name, "size") == 0)
+    if (status && strcmp (name, "hsize") == 0)
 	cubeUpdateGeometry (s, s->hsize, cs->invert);
 
     return status;
@@ -1715,6 +1936,10 @@ cubeInitScreen (CompPlugin *p,
 
     cs->unfoldVelocity = 0.0f;
 
+    cs->fullscreenOutput = TRUE;
+
+    memset (cs->cleared, 0, sizeof (cs->cleared));
+
     cubeScreenInitOptions (cs, s->display->display);
 
     cs->imgFiles = cs->opt[CUBE_SCREEN_OPTION_IMAGES].value.list.value;
@@ -1726,8 +1951,11 @@ cubeInitScreen (CompPlugin *p,
     WRAP (cs, s, donePaintScreen, cubeDonePaintScreen);
     WRAP (cs, s, paintScreen, cubePaintScreen);
     WRAP (cs, s, paintTransformedScreen, cubePaintTransformedScreen);
-    WRAP (cs, s, paintBackground, cubePaintBackground);
+    WRAP (cs, s, applyScreenTransform, cubeApplyScreenTransform);
     WRAP (cs, s, setScreenOption, cubeSetGlobalScreenOption);
+    WRAP (cs, s, outputChangeNotify, cubeOutputChangeNotify);
+
+    cubeUpdateOutputs (s);
 
     if (!cubeUpdateGeometry (s, s->hsize, cs->invert))
 	return FALSE;
@@ -1754,8 +1982,9 @@ cubeFiniScreen (CompPlugin *p,
     UNWRAP (cs, s, donePaintScreen);
     UNWRAP (cs, s, paintScreen);
     UNWRAP (cs, s, paintTransformedScreen);
-    UNWRAP (cs, s, paintBackground);
+    UNWRAP (cs, s, applyScreenTransform);
     UNWRAP (cs, s, setScreenOption);
+    UNWRAP (cs, s, outputChangeNotify);
 
     finiTexture (s, &cs->texture);
     finiTexture (s, &cs->sky);
