@@ -312,16 +312,21 @@ decorDisplayInitOptions (DecorDisplay *dd)
 }
 
 static Bool
-decorDrawWindow (CompWindow		 *w,
-		 const WindowPaintAttrib *attrib,
-		 Region			 region,
-		 unsigned int		 mask)
+decorDrawWindow (CompWindow	      *w,
+		 const CompTransform  *transform,
+		 const FragmentAttrib *attrib,
+		 Region		      region,
+		 unsigned int	      mask)
 {
     Bool status;
 
     DECOR_SCREEN (w->screen);
 
-    if (!(mask & PAINT_WINDOW_SOLID_MASK))
+    UNWRAP (ds, w->screen, drawWindow);
+    status = (*w->screen->drawWindow) (w, transform, attrib, region, mask);
+    WRAP (ds, w->screen, drawWindow, decorDrawWindow);
+
+    if (!(mask & PAINT_WINDOW_CLIP_OPAQUE_MASK))
     {
 	DECOR_WINDOW (w);
 
@@ -334,10 +339,12 @@ decorDrawWindow (CompWindow		 *w,
 	    REGION	     box;
 	    int		     i;
 
+	    mask |= PAINT_WINDOW_BLEND_MASK;
+
 	    box.rects	 = &box.extents;
 	    box.numRects = 1;
 
-	    w->vCount = 0;
+	    w->vCount = w->indexCount = 0;
 
 	    for (i = 0; i < wd->nQuad; i++)
 	    {
@@ -356,15 +363,10 @@ decorDrawWindow (CompWindow		 *w,
 	    if (w->vCount)
 		(*w->screen->drawWindowTexture) (w,
 						 &wd->decor->texture->texture,
-						 attrib, mask |
-						 PAINT_WINDOW_TRANSLUCENT_MASK);
+						 attrib, mask);
 	}
     }
 
-    UNWRAP (ds, w->screen, drawWindow);
-    status = (*w->screen->drawWindow) (w, attrib, region, mask);
-    WRAP (ds, w->screen, drawWindow, decorDrawWindow);
-
     return status;
 }
 
@@ -457,50 +459,6 @@ decorReleaseTexture (CompScreen   *screen,
     free (texture);
 }
 
-static void
-applyGravity (int gravity,
-	      int x,
-	      int y,
-	      int width,
-	      int height,
-	      int *return_x,
-	      int *return_y)
-{
-    if (gravity & GRAVITY_EAST)
-    {
-	x += width;
-	*return_x = MAX (0, x);
-    }
-    else if (gravity & GRAVITY_WEST)
-    {
-	*return_x = MIN (width, x);
-    }
-    else
-    {
-	x += width / 2;
-	x = MAX (0, x);
-	x = MIN (width, x);
-	*return_x = x;
-    }
-
-    if (gravity & GRAVITY_SOUTH)
-    {
-	y += height;
-	*return_y = MAX (0, y);
-    }
-    else if (gravity & GRAVITY_NORTH)
-    {
-	*return_y = MIN (height, y);
-    }
-    else
-    {
-	y += height / 2;
-	y = MAX (0, y);
-	y = MIN (height, y);
-	*return_y = y;
-    }
-}
-
 static void
 computeQuadBox (decor_quad_t *q,
 		int	     width,
@@ -512,8 +470,10 @@ computeQuadBox (decor_quad_t *q,
 {
     int x1, y1, x2, y2;
 
-    applyGravity (q->p1.gravity, q->p1.x, q->p1.y, width, height, &x1, &y1);
-    applyGravity (q->p2.gravity, q->p2.x, q->p2.y, width, height, &x2, &y2);
+    decor_apply_gravity (q->p1.gravity, q->p1.x, q->p1.y, width, height,
+			 &x1, &y1);
+    decor_apply_gravity (q->p2.gravity, q->p2.x, q->p2.y, width, height,
+			 &x2, &y2);
 
     if (q->clamp & CLAMP_HORZ)
     {
@@ -1007,16 +967,12 @@ static void
 decorHandleEvent (CompDisplay *d,
 		  XEvent      *event)
 {
-    Window     activeWindow = 0;
+    Window     activeWindow = d->activeWindow;
     CompWindow *w;
 
     DECOR_DISPLAY (d);
 
     switch (event->type) {
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	    activeWindow = d->activeWindow;
-	break;
     case DestroyNotify:
 	w = findWindowAtDisplay (d, event->xdestroywindow.window);
 	if (w)
@@ -1068,22 +1024,20 @@ decorHandleEvent (CompDisplay *d,
     (*d->handleEvent) (d, event);
     WRAP (dd, d, handleEvent, decorHandleEvent);
 
+    if (d->activeWindow != activeWindow)
+    {
+	w = findWindowAtDisplay (d, activeWindow);
+	if (w)
+	    decorWindowUpdate (w, FALSE);
+
+	w = findWindowAtDisplay (d, d->activeWindow);
+	if (w)
+	    decorWindowUpdate (w, FALSE);
+    }
+
     switch (event->type) {
     case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    if (d->activeWindow != activeWindow)
-	    {
-		w = findWindowAtDisplay (d, activeWindow);
-		if (w)
-		    decorWindowUpdate (w, FALSE);
-
-		w = findWindowAtDisplay (d, d->activeWindow);
-		if (w)
-		    decorWindowUpdate (w, FALSE);
-	    }
-	}
-	else if (event->xproperty.atom == dd->winDecorAtom)
+	if (event->xproperty.atom == dd->winDecorAtom)
 	{
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
