@@ -69,12 +69,12 @@
 
 #define ROTATE_SNAP_TOP_DEFAULT FALSE
 
-#define ROTATE_SPEED_DEFAULT   1.5f
+#define ROTATE_SPEED_DEFAULT   2.0f
 #define ROTATE_SPEED_MIN       0.1f
 #define ROTATE_SPEED_MAX       50.0f
 #define ROTATE_SPEED_PRECISION 0.1f
 
-#define ROTATE_TIMESTEP_DEFAULT   1.2f
+#define ROTATE_TIMESTEP_DEFAULT   1.0f
 #define ROTATE_TIMESTEP_MIN       0.1f
 #define ROTATE_TIMESTEP_MAX       50.0f
 #define ROTATE_TIMESTEP_PRECISION 0.1f
@@ -557,6 +557,7 @@ rotateDonePaintScreen (CompScreen *s)
 static Bool
 rotatePaintScreen (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   int			   output,
 		   unsigned int		   mask)
@@ -576,13 +577,14 @@ rotatePaintScreen (CompScreen		   *s,
 	mask |= PAINT_SCREEN_TRANSFORMED_MASK;
 
 	UNWRAP (rs, s, paintScreen);
-	status = (*s->paintScreen) (s, &sa, region, output, mask);
+	status = (*s->paintScreen) (s, &sa, transform, region, output, mask);
 	WRAP (rs, s, paintScreen, rotatePaintScreen);
     }
     else
     {
 	UNWRAP (rs, s, paintScreen);
-	status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+	status = (*s->paintScreen) (s, sAttrib, transform, region, output,
+				    mask);
 	WRAP (rs, s, paintScreen, rotatePaintScreen);
     }
 
@@ -606,6 +608,9 @@ rotateInitiate (CompDisplay     *d,
     {
 	ROTATE_SCREEN (s);
 
+	if (s->hsize < 2)
+	    return FALSE;
+
 	if (rs->rotateHandle && rs->grabWindow)
 	{
 	    if (otherScreenGrabExist (s, "rotate", "move", 0))
@@ -650,7 +655,7 @@ rotateInitiate (CompDisplay     *d,
 	}
     }
 
-    return FALSE;
+    return TRUE;
 }
 
 static Bool
@@ -706,6 +711,9 @@ rotate (CompDisplay     *d,
 
 	ROTATE_SCREEN (s);
 
+	if (s->hsize < 2)
+	    return FALSE;
+
 	if (otherScreenGrabExist (s, "rotate", "move", "switcher", "cube", 0))
 	    return FALSE;
 
@@ -766,6 +774,9 @@ rotateWithWindow (CompDisplay     *d,
 
 	ROTATE_SCREEN (s);
 
+	if (s->hsize < 2)
+	    return FALSE;
+
 	direction = getIntOptionNamed (option, nOption, "direction", 0);
 	if (!direction)
 	    return FALSE;
@@ -778,16 +789,19 @@ rotateWithWindow (CompDisplay     *d,
 
 	    rotateReleaseMoveWindow (s);
 
-	    w = findWindowAtScreen (s, xid);
-	    if (w)
+	    if (!rs->grabIndex && !rs->moving)
 	    {
-		if (!(w->type & (CompWindowTypeDesktopMask |
-				 CompWindowTypeDockMask)))
+		w = findWindowAtScreen (s, xid);
+		if (w)
 		{
-		    if (!(w->state & CompWindowStateStickyMask))
+		    if (!(w->type & (CompWindowTypeDesktopMask |
+				     CompWindowTypeDockMask)))
 		    {
-			rs->moveWindow  = w->id;
-			rs->moveWindowX = w->attrib.x;
+			if (!(w->state & CompWindowStateStickyMask))
+			{
+			    rs->moveWindow  = w->id;
+			    rs->moveWindowX = w->attrib.x;
+			}
 		    }
 		}
 	    }
@@ -1056,6 +1070,9 @@ rotateEdgeFlip (CompScreen      *s,
 
     ROTATE_DISPLAY (s->display);
 
+    if (s->hsize < 2)
+	return;
+
     if (otherScreenGrabExist (s, "rotate", "move", 0))
 	return;
 
@@ -1471,7 +1488,6 @@ rotateHandleEvent (CompDisplay *d,
 		    break;
 
 		/* reset movement */
-		rs->moving = FALSE;
 		rs->moveTo = 0.0f;
 
 		defaultViewportForWindow (w, &dx, NULL);
@@ -1518,9 +1534,14 @@ rotateHandleEvent (CompDisplay *d,
 	    {
 		int dx;
 
+		ROTATE_SCREEN (s);
+
 		if (otherScreenGrabExist (s, "rotate", "switcher", "cube", 0))
 		    break;
 
+		/* reset movement */
+		rs->moveTo = 0.0f;
+
 		dx = event->xclient.data.l[0] / s->width - s->x;
 		if (dx)
 		{
@@ -2136,7 +2157,9 @@ CompPluginVTable rotateVTable = {
     rotateGetScreenOptions,
     rotateSetScreenOption,
     rotateDeps,
-    sizeof (rotateDeps) / sizeof (rotateDeps[0])
+    sizeof (rotateDeps) / sizeof (rotateDeps[0]),
+    0, /* Features */
+    0  /* nFeatures */
 };
 
 CompPluginVTable *
