@@ -93,13 +93,18 @@ static char *winType[] = {
 };
 #define N_WIN_TYPE (sizeof (winType) / sizeof (winType[0]))
 
+#define ZOOMED_WINDOW_MASK (1 << 0)
+#define NORMAL_WINDOW_MASK (1 << 1)
+
 static int displayPrivateIndex;
 
 #define SWITCH_DISPLAY_OPTION_NEXT	     0
 #define SWITCH_DISPLAY_OPTION_PREV	     1
 #define SWITCH_DISPLAY_OPTION_NEXT_ALL	     2
 #define SWITCH_DISPLAY_OPTION_PREV_ALL	     3
-#define SWITCH_DISPLAY_OPTION_NUM	     4
+#define SWITCH_DISPLAY_OPTION_NEXT_NO_POPUP  4
+#define SWITCH_DISPLAY_OPTION_PREV_NO_POPUP  5
+#define SWITCH_DISPLAY_OPTION_NUM	     6
 
 typedef struct _SwitchDisplay {
     int		    screenPrivateIndex;
@@ -148,6 +153,7 @@ typedef struct _SwitchScreen {
 
     Bool switching;
     Bool zooming;
+    int  zoomMask;
 
     int moreAdjust;
 
@@ -518,7 +524,7 @@ compareWindows (const void *elem1,
 
     if (w2->mapNum && !w1->mapNum)
 	return 1;
-    
+
     return w2->activeNum - w1->activeNum;
 }
 
@@ -726,7 +732,8 @@ findArgbVisual (Display *dpy, int scr)
 
 static void
 switchInitiate (CompScreen *s,
-		Bool	   allWindows)
+		Bool	   allWindows,
+		Bool	   showPopup)
 {
     int count;
 
@@ -741,7 +748,7 @@ switchInitiate (CompScreen *s,
     if (count < 1)
 	return;
 
-    if (!ss->popupWindow)
+    if (!ss->popupWindow && showPopup)
     {
 	Display		     *dpy = s->display->display;
 	XSizeHints	     xsh;
@@ -812,6 +819,10 @@ switchInitiate (CompScreen *s,
 	setWindowProp (s->display, ss->popupWindow,
 		       s->display->winDesktopAtom,
 		       0xffffffff);
+
+	setWindowProp (s->display, ss->popupWindow,
+		       s->display->winTypeAtom,
+		       s->display->winTypeUtilAtom);
     }
 
     if (!ss->grabIndex)
@@ -827,7 +838,7 @@ switchInitiate (CompScreen *s,
 
 	    ss->sTranslate = ss->zoom;
 
-	    if (ss->popupWindow)
+	    if (ss->popupWindow && showPopup)
 	    {
 		CompWindow *w;
 
@@ -943,13 +954,17 @@ switchNext (CompDisplay     *d,
 
 	if (!ss->switching)
 	{
-	    switchInitiate (s, FALSE);
+	    switchInitiate (s, FALSE, TRUE);
 
 	    if (state & CompActionStateInitKey)
 		action->state |= CompActionStateTermKey;
 
 	    if (state & CompActionStateInitButton)
 		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
+
 	}
 
 	switchToWindow (s, TRUE);
@@ -977,13 +992,16 @@ switchPrev (CompDisplay     *d,
 
 	if (!ss->switching)
 	{
-	    switchInitiate (s, FALSE);
+	    switchInitiate (s, FALSE, TRUE);
 
 	    if (state & CompActionStateInitKey)
 		action->state |= CompActionStateTermKey;
 
 	    if (state & CompActionStateInitButton)
 		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
 	}
 
 	switchToWindow (s, FALSE);
@@ -1011,13 +1029,16 @@ switchNextAll (CompDisplay     *d,
 
 	if (!ss->switching)
 	{
-	    switchInitiate (s, TRUE);
+	    switchInitiate (s, TRUE, TRUE);
 
 	    if (state & CompActionStateInitKey)
 		action->state |= CompActionStateTermKey;
 
 	    if (state & CompActionStateInitButton)
 		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
 	}
 
 	switchToWindow (s, TRUE);
@@ -1045,13 +1066,91 @@ switchPrevAll (CompDisplay     *d,
 
 	if (!ss->switching)
 	{
-	    switchInitiate (s, TRUE);
+	    switchInitiate (s, TRUE, TRUE);
 
 	    if (state & CompActionStateInitKey)
 		action->state |= CompActionStateTermKey;
 
 	    if (state & CompActionStateInitButton)
 		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
+	}
+
+	switchToWindow (s, FALSE);
+    }
+
+    return FALSE;
+}
+
+static Bool
+switchNextNoPopup (CompDisplay     *d,
+		   CompAction      *action,
+		   CompActionState state,
+		   CompOption      *option,
+		   int	           nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+    {
+	SWITCH_SCREEN (s);
+
+	if (!ss->switching)
+	{
+	    switchInitiate (s, FALSE, FALSE);
+
+	    if (state & CompActionStateInitKey)
+		action->state |= CompActionStateTermKey;
+
+	    if (state & CompActionStateInitButton)
+		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
+
+	}
+
+	switchToWindow (s, TRUE);
+    }
+
+    return FALSE;
+}
+
+static Bool
+switchPrevNoPopup (CompDisplay     *d,
+		   CompAction      *action,
+		   CompActionState state,
+		   CompOption      *option,
+		   int	           nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+    {
+	SWITCH_SCREEN (s);
+
+	if (!ss->switching)
+	{
+	    switchInitiate (s, FALSE, FALSE);
+
+	    if (state & CompActionStateInitKey)
+		action->state |= CompActionStateTermKey;
+
+	    if (state & CompActionStateInitButton)
+		action->state |= CompActionStateTermButton;
+
+	    if (state & CompActionStateInitEdge)
+		action->state |= CompActionStateTermEdge;
 	}
 
 	switchToWindow (s, FALSE);
@@ -1250,8 +1349,8 @@ adjustSwitchVelocity (CompScreen *s)
 	if (ss->selectedWindow == ss->zoomedWindow)
 	{
 	    if (fabs (dx) < 0.1f   && fabs (ss->mVelocity) < 0.2f   &&
-		fabs (dt) < 0.002f && fabs (ss->tVelocity) < 0.002f &&
-		fabs (ds) < 0.002f && fabs (ss->sVelocity) < 0.002f)
+		fabs (dt) < 0.001f && fabs (ss->tVelocity) < 0.001f &&
+		fabs (ds) < 0.001f && fabs (ss->sVelocity) < 0.001f)
 	    {
 		ss->mVelocity = ss->tVelocity = ss->sVelocity = 0.0f;
 		return 0;
@@ -1352,6 +1451,7 @@ switchPreparePaintScreen (CompScreen *s,
 static Bool
 switchPaintScreen (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   int			   output,
 		   unsigned int		   mask)
@@ -1360,6 +1460,8 @@ switchPaintScreen (CompScreen		   *s,
 
     SWITCH_SCREEN (s);
 
+    ss->zoomMask = ZOOMED_WINDOW_MASK | NORMAL_WINDOW_MASK;
+
     if (ss->grabIndex || (ss->zooming && ss->translate > 0.001f))
     {
 	ScreenPaintAttrib sa = *sAttrib;
@@ -1374,6 +1476,8 @@ switchPaintScreen (CompScreen		   *s,
 	    mask |= PAINT_SCREEN_TRANSFORMED_MASK | PAINT_SCREEN_CLEAR_MASK;
 
 	    sa.zCamera -= ss->translate;
+
+	    ss->zoomMask = NORMAL_WINDOW_MASK;
 	}
 
 	switcher = findWindowAtScreen (s, ss->popupWindow);
@@ -1403,9 +1507,23 @@ switchPaintScreen (CompScreen		   *s,
 	}
 
 	UNWRAP (ss, s, paintScreen);
-	status = (*s->paintScreen) (s, &sa, region, output, mask);
+	status = (*s->paintScreen) (s, &sa, transform, region, output, mask);
 	WRAP (ss, s, paintScreen, switchPaintScreen);
 
+	if (ss->zooming)
+	{
+	    mask &= ~PAINT_SCREEN_CLEAR_MASK;
+
+	    ss->zoomMask = ZOOMED_WINDOW_MASK;
+
+	    sa.zCamera += MIN (ss->sTranslate, ss->translate);
+
+	    UNWRAP (ss, s, paintScreen);
+	    status = (*s->paintScreen) (s, &sa, transform, region, output,
+					mask);
+	    WRAP (ss, s, paintScreen, switchPaintScreen);
+	}
+
 	if (zoomed)
 	{
 	    unhookWindowFromScreen (s, zoomed);
@@ -1414,17 +1532,20 @@ switchPaintScreen (CompScreen		   *s,
 
 	if (switcher)
 	{
+	    CompTransform sTransform = *transform;
+
 	    switcher->destroyed = saveDestroyed;
 
-	    glPushMatrix ();
+	    transformToScreenSpace (s, output, -DEFAULT_Z_CAMERA, &sTransform);
 
-	    prepareXCoords (s, output, -DEFAULT_Z_CAMERA);
+	    glPushMatrix ();
+	    glLoadMatrixf (sTransform.m);
 
 	    if (!switcher->destroyed			 &&
 		switcher->attrib.map_state == IsViewable &&
 		switcher->damaged)
 	    {
-		(*s->paintWindow) (switcher, &switcher->paint,
+		(*s->paintWindow) (switcher, &switcher->paint, &sTransform,
 				   &infiniteRegion, 0);
 	    }
 
@@ -1434,7 +1555,8 @@ switchPaintScreen (CompScreen		   *s,
     else
     {
 	UNWRAP (ss, s, paintScreen);
-	status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+	status = (*s->paintScreen) (s, sAttrib, transform, region, output,
+				    mask);
 	WRAP (ss, s, paintScreen, switchPaintScreen);
     }
 
@@ -1467,68 +1589,51 @@ switchDonePaintScreen (CompScreen *s)
     WRAP (ss, s, donePaintScreen, switchDonePaintScreen);
 }
 
-static void
-switchMoveWindow (CompWindow *w,
-		  int        dx,
-		  int        dy)
-{
-    w->attrib.x += dx;
-    w->attrib.y += dy;
-
-    XOffsetRegion (w->region, dx, dy);
-}
-
 static void
 switchPaintThumb (CompWindow		  *w,
 		  const WindowPaintAttrib *attrib,
+		  const CompTransform	  *transform,
 		  unsigned int		  mask,
 		  int			  x,
 		  int			  y,
 		  int			  x1,
 		  int			  x2)
 {
-    DrawWindowGeometryProc oldDrawWindowGeometry;
-    WindowPaintAttrib	   sAttrib = *attrib;
-    int			   dx, dy;
-    int			   wx, wy;
-    float		   width, height;
-    REGION		   reg;
-    CompMatrix		   matrix;
-    CompIcon		   *icon = NULL;
-
-    /* Wrap drawWindowGeometry to make sure the general
-       drawWindowGeometry function is used */
-    oldDrawWindowGeometry = w->screen->drawWindowGeometry;
-    w->screen->drawWindowGeometry = drawWindowGeometry;
+    WindowPaintAttrib sAttrib = *attrib;
+    int		      wx, wy;
+    float	      width, height;
+    CompIcon	      *icon = NULL;
 
     mask |= PAINT_WINDOW_TRANSFORMED_MASK;
 
-    reg.rects    = &reg.extents;
-    reg.numRects = 1;
-
-    reg.extents.y1 = MINSHORT;
-    reg.extents.y2 = MAXSHORT;
-
     if (w->mapNum)
     {
-	if (!w->texture->pixmap)
+	if (!w->texture->pixmap && !w->bindFailed)
 	    bindWindow (w);
     }
 
-    if (w->mapNum)
+    if (w->texture->pixmap)
     {
+	AddWindowGeometryProc oldAddWindowGeometry;
+	FragmentAttrib	      fragment;
+	CompTransform	      wTransform = *transform;
+	int		      ww, wh;
+
 	SWITCH_SCREEN (w->screen);
 
 	width  = WIDTH  - (SPACE << 1);
 	height = HEIGHT - (SPACE << 1);
 
-	if (w->width > width)
-	    sAttrib.xScale = width / w->width;
+	ww = w->width  + w->input.left + w->input.right;
+	wh = w->height + w->input.top  + w->input.bottom;
+
+	if (ww > width)
+	    sAttrib.xScale = width / ww;
 	else
 	    sAttrib.xScale = 1.0f;
 
-	if (w->height > height)
-	    sAttrib.yScale = height / w->height;
+	if (wh > height)
+	    sAttrib.yScale = height / wh;
 	else
 	    sAttrib.yScale = 1.0f;
 
@@ -1537,35 +1642,40 @@ switchPaintThumb (CompWindow		  *w,
 	else
 	    sAttrib.xScale = sAttrib.yScale;
 
-	width  = w->width  * sAttrib.xScale;
-	height = w->height * sAttrib.yScale;
+	width  = ww * sAttrib.xScale;
+	height = wh * sAttrib.yScale;
 
 	wx = x + SPACE + ((WIDTH  - (SPACE << 1)) - width)  / 2;
 	wy = y + SPACE + ((HEIGHT - (SPACE << 1)) - height) / 2;
 
-	dx = wx - w->attrib.x;
-	dy = wy - w->attrib.y;
+	sAttrib.xTranslate = wx - w->attrib.x + w->input.left * sAttrib.xScale;
+	sAttrib.yTranslate = wy - w->attrib.y + w->input.top  * sAttrib.yScale;
 
-	switchMoveWindow (w, dx, dy);
+	initFragmentAttrib (&fragment, &sAttrib);
 
-	matrix = w->texture->matrix;
-	matrix.x0 -= (w->attrib.x * w->matrix.xx);
-	matrix.y0 -= (w->attrib.y * w->matrix.yy);
-
-	if (w->alpha || sAttrib.opacity != OPAQUE)
+	if (w->alpha || fragment.opacity != OPAQUE)
 	    mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
-	else if (sAttrib.opacity == OPAQUE)
-	    mask &= ~PAINT_WINDOW_TRANSLUCENT_MASK;
 
-	reg.extents.x1 = wx + (x1 - wx) / sAttrib.xScale;
-	reg.extents.x2 = wx + (x2 - wx) / sAttrib.xScale;
+	matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
+	matrixScale (&wTransform, sAttrib.xScale, sAttrib.yScale, 0.0f);
+	matrixTranslate (&wTransform,
+			 sAttrib.xTranslate / sAttrib.xScale - w->attrib.x,
+			 sAttrib.yTranslate / sAttrib.yScale - w->attrib.y,
+			 0.0f);
 
-	w->vCount = 0;
-	addWindowGeometry (w, &matrix, 1, w->region, &reg);
-	if (w->vCount)
-	    (*w->screen->drawWindowTexture) (w, w->texture, &sAttrib, mask);
+	glPushMatrix ();
+	glLoadMatrixf (wTransform.m);
 
-	switchMoveWindow (w, -dx, -dy);
+	/* XXX: replacing the addWindowGeometry function like this is
+	   very ugly but necessary until the vertex stage has been made
+	   fully pluggable. */
+	oldAddWindowGeometry = w->screen->addWindowGeometry;
+	w->screen->addWindowGeometry = addWindowGeometry;
+	(w->screen->drawWindow) (w, &wTransform, &fragment, &infiniteRegion,
+				 mask);
+	w->screen->addWindowGeometry = oldAddWindowGeometry;
+
+	glPopMatrix ();
 
 	if (ss->opt[SWITCH_SCREEN_OPTION_ICON].value.b)
 	{
@@ -1620,14 +1730,10 @@ switchPaintThumb (CompWindow		  *w,
 
     if (icon && (icon->texture.name || iconToTexture (w->screen, icon)))
     {
-	REGION iconReg;
+	REGION     iconReg;
+	CompMatrix matrix;
 
-	mask |= PAINT_WINDOW_TRANSLUCENT_MASK;
-
-	dx = wx - w->attrib.x;
-	dy = wy - w->attrib.y;
-
-	switchMoveWindow (w, dx, dy);
+	mask |= PAINT_WINDOW_BLEND_MASK;
 
 	iconReg.rects    = &iconReg.extents;
 	iconReg.numRects = 1;
@@ -1638,32 +1744,49 @@ switchPaintThumb (CompWindow		  *w,
 	iconReg.extents.y2 = w->attrib.y + icon->height;
 
 	matrix = icon->texture.matrix;
-	matrix.x0 -= (wx * icon->texture.matrix.xx);
-	matrix.y0 -= (wy * icon->texture.matrix.yy);
+	matrix.x0 -= (w->attrib.x * icon->texture.matrix.xx);
+	matrix.y0 -= (w->attrib.y * icon->texture.matrix.yy);
 
-	reg.extents.x1 = wx + (x1 - wx) / sAttrib.xScale;
-	reg.extents.x2 = wx + (x2 - wx) / sAttrib.xScale;
+	sAttrib.xTranslate = wx - w->attrib.x;
+	sAttrib.yTranslate = wy - w->attrib.y;
 
-	w->vCount = 0;
-	addWindowGeometry (w, &matrix, 1, &iconReg, &reg);
+	w->vCount = w->indexCount = 0;
+	addWindowGeometry (w, &matrix, 1, &iconReg, &infiniteRegion);
 	if (w->vCount)
+	{
+	    FragmentAttrib fragment;
+	    CompTransform  wTransform = *transform;
+
+	    initFragmentAttrib (&fragment, &sAttrib);
+
+	    matrixTranslate (&wTransform, w->attrib.x, w->attrib.y, 0.0f);
+	    matrixScale (&wTransform, sAttrib.xScale, sAttrib.yScale, 0.0f);
+	    matrixTranslate (&wTransform,
+			     sAttrib.xTranslate / sAttrib.xScale - w->attrib.x,
+			     sAttrib.yTranslate / sAttrib.yScale - w->attrib.y,
+			     0.0f);
+
+	    glPushMatrix ();
+	    glLoadMatrixf (wTransform.m);
+
 	    (*w->screen->drawWindowTexture) (w,
-					     &icon->texture, &sAttrib,
+					     &icon->texture, &fragment,
 					     mask);
 
-	switchMoveWindow (w, -dx, -dy);
+	    glPopMatrix ();
+	}
     }
-
-    w->screen->drawWindowGeometry = oldDrawWindowGeometry;
 }
 
 static Bool
 switchPaintWindow (CompWindow		   *w,
 		   const WindowPaintAttrib *attrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   unsigned int		   mask)
 {
     CompScreen *s = w->screen;
+    int	       zoomType = NORMAL_WINDOW_MASK;
     Bool       status;
 
     SWITCH_SCREEN (s);
@@ -1673,11 +1796,11 @@ switchPaintWindow (CompWindow		   *w,
 	GLenum filter;
 	int    x, y, x1, x2, cx, i;
 
-	if (mask & PAINT_WINDOW_SOLID_MASK)
+	if (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK)
 	    return FALSE;
 
 	UNWRAP (ss, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (ss, s, paintWindow, switchPaintWindow);
 
 	if (!(mask & PAINT_WINDOW_TRANSFORMED_MASK) && region->numRects == 0)
@@ -1694,11 +1817,16 @@ switchPaintWindow (CompWindow		   *w,
 	if (ss->opt[SWITCH_SCREEN_OPTION_MIPMAP].value.b)
 	    s->display->textureFilter = GL_LINEAR_MIPMAP_LINEAR;
 
+	glPushAttrib (GL_SCISSOR_BIT);
+
+	glEnable (GL_SCISSOR_TEST);
+	glScissor (x1, 0, x2 - x1, w->screen->height);
+
 	for (i = 0; i < ss->nWindows; i++)
 	{
 	    if (x + WIDTH > x1)
-		switchPaintThumb (ss->windows[i], &w->lastPaint, mask,
-				  x, y, x1, x2);
+		switchPaintThumb (ss->windows[i], &w->lastPaint, transform,
+				  mask, x, y, x1, x2);
 
 	    x += WIDTH;
 	}
@@ -1708,12 +1836,14 @@ switchPaintWindow (CompWindow		   *w,
 	    if (x > x2)
 		break;
 
-	    switchPaintThumb (ss->windows[i], &w->lastPaint, mask,
+	    switchPaintThumb (ss->windows[i], &w->lastPaint, transform, mask,
 			      x, y, x1, x2);
 
 	    x += WIDTH;
 	}
 
+	glPopAttrib ();
+
 	s->display->textureFilter = filter;
 
 	cx = w->attrib.x + (w->width >> 1);
@@ -1732,24 +1862,15 @@ switchPaintWindow (CompWindow		   *w,
     }
     else if (w->id == ss->selectedWindow)
     {
-	glPushMatrix ();
+	if (ss->bringToFront && ss->selectedWindow == ss->zoomedWindow)
+	    zoomType = ZOOMED_WINDOW_MASK;
 
-	if (ss->bringToFront)
-	{
-	    if (ss->selectedWindow == ss->zoomedWindow)
-		glTranslatef (0.0f, 0.0f, MIN (ss->translate, ss->sTranslate));
-	}
-
-	glPushAttrib (GL_STENCIL_BUFFER_BIT);
-	glDisable (GL_STENCIL_TEST);
+	if (!(ss->zoomMask & zoomType))
+	    return (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK) ? FALSE : TRUE;
 
 	UNWRAP (ss, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (ss, s, paintWindow, switchPaintWindow);
-
-	glPopAttrib ();
-
-	glPopMatrix ();
     }
     else if (ss->switching)
     {
@@ -1765,32 +1886,22 @@ switchPaintWindow (CompWindow		   *w,
 	    sAttrib.opacity = (sAttrib.opacity * ss->opacity) >> 16;
 
 	if (ss->bringToFront && w->id == ss->zoomedWindow)
-	{
-	    glPushMatrix ();
-	    glTranslatef (0.0f, 0.0f, MIN (ss->translate, ss->sTranslate));
+	    zoomType = ZOOMED_WINDOW_MASK;
 
-	    glPushAttrib (GL_STENCIL_BUFFER_BIT);
-	    glDisable (GL_STENCIL_TEST);
+	if (!(ss->zoomMask & zoomType))
+	    return (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK) ? FALSE : TRUE;
 
-	    UNWRAP (ss, s, paintWindow);
-	    status = (*s->paintWindow) (w, &sAttrib, region, mask);
-	    WRAP (ss, s, paintWindow, switchPaintWindow);
-
-	    glPopAttrib ();
-
-	    glPopMatrix ();
-	}
-	else
-	{
-	    UNWRAP (ss, s, paintWindow);
-	    status = (*s->paintWindow) (w, &sAttrib, region, mask);
-	    WRAP (ss, s, paintWindow, switchPaintWindow);
-	}
+	UNWRAP (ss, s, paintWindow);
+	status = (*s->paintWindow) (w, &sAttrib, transform, region, mask);
+	WRAP (ss, s, paintWindow, switchPaintWindow);
     }
     else
     {
+	if (!(ss->zoomMask & zoomType))
+	    return (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK) ? FALSE : TRUE;
+
 	UNWRAP (ss, s, paintWindow);
-	status = (*s->paintWindow) (w, attrib, region, mask);
+	status = (*s->paintWindow) (w, attrib, transform, region, mask);
 	WRAP (ss, s, paintWindow, switchPaintWindow);
     }
 
@@ -1858,8 +1969,10 @@ switchSetDisplayOption (CompDisplay     *display,
     switch (index) {
     case SWITCH_DISPLAY_OPTION_NEXT:
     case SWITCH_DISPLAY_OPTION_NEXT_ALL:
+    case SWITCH_DISPLAY_OPTION_NEXT_NO_POPUP:
     case SWITCH_DISPLAY_OPTION_PREV:
     case SWITCH_DISPLAY_OPTION_PREV_ALL:
+    case SWITCH_DISPLAY_OPTION_PREV_NO_POPUP:
 	if (setDisplayAction (display, o, value))
 	    return TRUE;
     default:
@@ -1948,6 +2061,32 @@ switchDisplayInitOptions (SwitchDisplay *sd,
     o->value.action.key.keycode   =
 	XKeysymToKeycode (display,
 			  XStringToKeysym (SWITCH_PREV_ALL_KEY_DEFAULT));
+
+    o = &sd->opt[SWITCH_DISPLAY_OPTION_NEXT_NO_POPUP];
+    o->name		      = "next_no_popup";
+    o->shortDesc	      = N_("Next window");
+    o->longDesc		      = N_("Select next window");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = switchNextNoPopup;
+    o->value.action.terminate = switchTerminate;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
+    o->value.action.type      = 0;
+
+    o = &sd->opt[SWITCH_DISPLAY_OPTION_PREV_NO_POPUP];
+    o->name		      = "prev_no_popup";
+    o->shortDesc	      = N_("Prev window");
+    o->longDesc		      = N_("Select previous window");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = switchPrevNoPopup;
+    o->value.action.terminate = switchTerminate;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
+    o->value.action.type      = 0;
 }
 
 static Bool
@@ -2026,6 +2165,8 @@ switchInitScreen (CompPlugin *p,
 
     ss->zooming  = (SWITCH_ZOOM_DEFAULT > 0.05f) ? TRUE : FALSE;
 
+    ss->zoomMask = ~0;
+
     ss->moreAdjust = 0;
 
     ss->mVelocity = 0.0f;
@@ -2048,6 +2189,8 @@ switchInitScreen (CompPlugin *p,
     addScreenAction (s, &sd->opt[SWITCH_DISPLAY_OPTION_PREV].value.action);
     addScreenAction (s, &sd->opt[SWITCH_DISPLAY_OPTION_NEXT_ALL].value.action);
     addScreenAction (s, &sd->opt[SWITCH_DISPLAY_OPTION_PREV_ALL].value.action);
+    addScreenAction (s, &sd->opt[SWITCH_DISPLAY_OPTION_NEXT_NO_POPUP].value.action);
+    addScreenAction (s, &sd->opt[SWITCH_DISPLAY_OPTION_PREV_NO_POPUP].value.action);
 
     WRAP (ss, s, preparePaintScreen, switchPreparePaintScreen);
     WRAP (ss, s, donePaintScreen, switchDonePaintScreen);
