@@ -98,9 +98,6 @@ typedef struct _Model {
     int		 numSprings;
     Object	 *anchorObject;
     float	 steps;
-    Vector	 scale;
-    Point	 scaleOrigin;
-    Bool	 transformed;
     Point	 topLeft;
     Point	 bottomRight;
     unsigned int edgeMask;
@@ -213,9 +210,6 @@ typedef struct _WobblyScreen {
     PaintWindowProc	   paintWindow;
     DamageWindowRectProc   damageWindowRect;
     AddWindowGeometryProc  addWindowGeometry;
-    DrawWindowGeometryProc drawWindowGeometry;
-    DrawWindowTextureProc  drawWindowTexture;
-    SetWindowScaleProc     setWindowScale;
 
     WindowResizeNotifyProc windowResizeNotify;
     WindowMoveNotifyProc   windowMoveNotify;
@@ -506,6 +500,11 @@ wobblyScreenInitOptions (WobblyScreen *ws,
     o->value.b    = WOBBLY_MAXIMIZE_EFFECT_DEFAULT;
 }
 
+#define SNAP_WINDOW_TYPE (CompWindowTypeNormalMask  | \
+			  CompWindowTypeToolbarMask | \
+			  CompWindowTypeMenuMask    | \
+			  CompWindowTypeUtilMask)
+
 static void
 findNextWestEdge (CompWindow *w,
 		  Object     *object)
@@ -514,6 +513,7 @@ findNextWestEdge (CompWindow *w,
     int s, start;
     int e, end;
     int x;
+    int output;
 
     start = -65535.0f;
     end   =  65535.0f;
@@ -523,19 +523,35 @@ findNextWestEdge (CompWindow *w,
 
     x = object->position.x + w->output.left - w->input.left;
 
-    if (x >= w->screen->workArea.x)
+    output = outputDeviceForPoint (w->screen, x, object->position.y);
+
+    if (x >= w->screen->outputDev[output].region.extents.x1)
     {
 	CompWindow *p;
 
-	v1 = w->screen->workArea.x;
+	v1 = w->screen->outputDev[output].region.extents.x1;
 
 	for (p = w->screen->windows; p; p = p->next)
 	{
-	    if (p->invisible || w == p || p->type != CompWindowTypeNormalMask)
+	    if (w == p)
 		continue;
 
-	    s = p->attrib.y - p->output.top;
-	    e = p->attrib.y + p->height + p->output.bottom;
+	    if (p->mapNum && p->struts)
+	    {
+		s = p->struts->left.y - w->output.top;
+		e = p->struts->left.y + p->struts->left.height +
+		    w->output.bottom;
+	    }
+	    else if (!p->invisible && (p->type & SNAP_WINDOW_TYPE))
+	    {
+		s = p->attrib.y - p->input.top - w->output.top;
+		e = p->attrib.y + p->height + p->input.bottom +
+		    w->output.bottom;
+	    }
+	    else
+	    {
+		continue;
+	    }
 
 	    if (s > object->position.y)
 	    {
@@ -555,7 +571,11 @@ findNextWestEdge (CompWindow *w,
 		if (e < end)
 		    end = e;
 
-		v = p->attrib.x + p->width + p->input.right;
+		if (p->mapNum && p->struts)
+		    v = p->struts->left.x + p->struts->left.width;
+		else
+		    v = p->attrib.x + p->width + p->input.right;
+
 		if (v <= x)
 		{
 		    if (v > v1)
@@ -571,7 +591,7 @@ findNextWestEdge (CompWindow *w,
     }
     else
     {
-	v2 = w->screen->workArea.x;
+	v2 = w->screen->outputDev[output].region.extents.x1;
     }
 
     v1 = v1 - w->output.left + w->input.left;
@@ -598,6 +618,7 @@ findNextEastEdge (CompWindow *w,
     int s, start;
     int e, end;
     int x;
+    int output;
 
     start = -65535.0f;
     end   =  65535.0f;
@@ -607,19 +628,35 @@ findNextEastEdge (CompWindow *w,
 
     x = object->position.x - w->output.right + w->input.right;
 
-    if (x <= w->screen->workArea.x + w->screen->workArea.width)
+    output = outputDeviceForPoint (w->screen, x, object->position.y);
+
+    if (x <= w->screen->outputDev[output].region.extents.x2)
     {
 	CompWindow *p;
 
-	v1 = w->screen->workArea.x + w->screen->workArea.width;
+	v1 = w->screen->outputDev[output].region.extents.x2;
 
 	for (p = w->screen->windows; p; p = p->next)
 	{
-	    if (p->invisible || w == p || p->type != CompWindowTypeNormalMask)
+	    if (w == p)
 		continue;
 
-	    s = p->attrib.y - p->output.top;
-	    e = p->attrib.y + p->height + p->output.bottom;
+	    if (p->mapNum && p->struts)
+	    {
+		s = p->struts->right.y - w->output.top;
+		e = p->struts->right.y + p->struts->right.height +
+		    w->output.bottom;
+	    }
+	    else if (!p->invisible && (p->type & SNAP_WINDOW_TYPE))
+	    {
+		s = p->attrib.y - p->input.top - w->output.top;
+		e = p->attrib.y + p->height + p->input.bottom +
+		    w->output.bottom;
+	    }
+	    else
+	    {
+		continue;
+	    }
 
 	    if (s > object->position.y)
 	    {
@@ -639,7 +676,11 @@ findNextEastEdge (CompWindow *w,
 		if (e < end)
 		    end = e;
 
-		v = p->attrib.x - p->input.left;
+		if (p->mapNum && p->struts)
+		    v = p->struts->right.x;
+		else
+		    v = p->attrib.x - p->input.left;
+
 		if (v >= x)
 		{
 		    if (v < v1)
@@ -655,7 +696,7 @@ findNextEastEdge (CompWindow *w,
     }
     else
     {
-	v2 = w->screen->workArea.x + w->screen->workArea.width;
+	v2 = w->screen->outputDev[output].region.extents.x2;
     }
 
     v1 = v1 + w->output.right - w->input.right;
@@ -682,6 +723,7 @@ findNextNorthEdge (CompWindow *w,
     int s, start;
     int e, end;
     int y;
+    int output;
 
     start = -65535.0f;
     end   =  65535.0f;
@@ -691,19 +733,33 @@ findNextNorthEdge (CompWindow *w,
 
     y = object->position.y + w->output.top - w->input.top;
 
-    if (y >= w->screen->workArea.y)
+    output = outputDeviceForPoint (w->screen, object->position.x, y);
+
+    if (y >= w->screen->outputDev[output].region.extents.y1)
     {
 	CompWindow *p;
 
-	v1 = w->screen->workArea.y;
+	v1 = w->screen->outputDev[output].region.extents.y1;
 
 	for (p = w->screen->windows; p; p = p->next)
 	{
-	    if (p->invisible || w == p || p->type != CompWindowTypeNormalMask)
+	    if (w == p)
 		continue;
 
-	    s = p->attrib.x - p->output.left;
-	    e = p->attrib.x + p->width + p->output.right;
+	    if (p->mapNum && p->struts)
+	    {
+		s = p->struts->top.x - w->output.left;
+		e = p->struts->top.x + p->struts->top.width + w->output.right;
+	    }
+	    else if (!p->invisible && (p->type & SNAP_WINDOW_TYPE))
+	    {
+		s = p->attrib.x - p->input.left - w->output.left;
+		e = p->attrib.x + p->width + p->input.right + w->output.right;
+	    }
+	    else
+	    {
+		continue;
+	    }
 
 	    if (s > object->position.x)
 	    {
@@ -723,7 +779,11 @@ findNextNorthEdge (CompWindow *w,
 		if (e < end)
 		    end = e;
 
-		v = p->attrib.y + p->height + p->input.bottom;
+		if (p->mapNum && p->struts)
+		    v = p->struts->top.y + p->struts->top.height;
+		else
+		    v = p->attrib.y + p->height + p->input.bottom;
+
 		if (v <= y)
 		{
 		    if (v > v1)
@@ -739,7 +799,7 @@ findNextNorthEdge (CompWindow *w,
     }
     else
     {
-	v2 = w->screen->workArea.y;
+	v2 = w->screen->outputDev[output].region.extents.y1;
     }
 
     v1 = v1 - w->output.top + w->input.top;
@@ -766,6 +826,7 @@ findNextSouthEdge (CompWindow *w,
     int s, start;
     int e, end;
     int y;
+    int output;
 
     start = -65535.0f;
     end   =  65535.0f;
@@ -775,19 +836,34 @@ findNextSouthEdge (CompWindow *w,
 
     y = object->position.y - w->output.bottom + w->input.bottom;
 
-    if (y <= w->screen->workArea.y + w->screen->workArea.height)
+    output = outputDeviceForPoint (w->screen, object->position.x, y);
+
+    if (y <= w->screen->outputDev[output].region.extents.y2)
     {
 	CompWindow *p;
 
-	v1 = w->screen->workArea.y + w->screen->workArea.height;
+	v1 = w->screen->outputDev[output].region.extents.y2;
 
 	for (p = w->screen->windows; p; p = p->next)
 	{
-	    if (p->invisible || w == p || p->type != CompWindowTypeNormalMask)
+	    if (w == p)
 		continue;
 
-	    s = p->attrib.x - p->output.left;
-	    e = p->attrib.x + p->width + p->output.right;
+	    if (p->mapNum && p->struts)
+	    {
+		s = p->struts->bottom.x - w->output.left;
+		e = p->struts->bottom.x + p->struts->bottom.width +
+		    w->output.right;
+	    }
+	    else if (!p->invisible && (p->type & SNAP_WINDOW_TYPE))
+	    {
+		s = p->attrib.x - p->input.left - w->output.left;
+		e = p->attrib.x + p->width + p->input.right + w->output.right;
+	    }
+	    else
+	    {
+		continue;
+	    }
 
 	    if (s > object->position.x)
 	    {
@@ -807,7 +883,11 @@ findNextSouthEdge (CompWindow *w,
 		if (e < end)
 		    end = e;
 
-		v = p->attrib.y - p->input.top;
+		if (p->mapNum && p->struts)
+		    v = p->struts->bottom.y;
+		else
+		    v = p->attrib.y - p->input.top;
+
 		if (v >= y)
 		{
 		    if (v < v1)
@@ -823,7 +903,7 @@ findNextSouthEdge (CompWindow *w,
     }
     else
     {
-	v2 = w->screen->workArea.y + w->screen->workArea.height;
+	v2 = w->screen->outputDev[output].region.extents.y2;
     }
 
     v1 = v1 + w->output.bottom - w->input.bottom;
@@ -929,10 +1009,7 @@ modelSetMiddleAnchor (Model *model,
 		      int   width,
 		      int   height)
 {
-    float gx, gy, x0, y0;
-
-    x0 = model->scaleOrigin.x;
-    y0 = model->scaleOrigin.y;
+    float gx, gy;
 
     gx = ((GRID_WIDTH  - 1) / 2 * width)  / (float) (GRID_WIDTH  - 1);
     gy = ((GRID_HEIGHT - 1) / 2 * height) / (float) (GRID_HEIGHT - 1);
@@ -943,8 +1020,28 @@ modelSetMiddleAnchor (Model *model,
     model->anchorObject = &model->objects[GRID_WIDTH *
 					  ((GRID_HEIGHT - 1) / 2) +
 					  (GRID_WIDTH - 1) / 2];
-    model->anchorObject->position.x = x + (gx - x0) * model->scale.x + x0;
-    model->anchorObject->position.y = y + (gy - y0) * model->scale.y + y0;
+    model->anchorObject->position.x = x + gx;
+    model->anchorObject->position.y = y + gy;
+
+    model->anchorObject->immobile = TRUE;
+}
+
+static void
+modelSetTopAnchor (Model *model,
+		   int   x,
+		   int   y,
+		   int   width)
+{
+    float gx;
+
+    gx = ((GRID_WIDTH - 1) / 2 * width)  / (float) (GRID_WIDTH - 1);
+
+    if (model->anchorObject)
+	model->anchorObject->immobile = FALSE;
+
+    model->anchorObject = &model->objects[(GRID_WIDTH - 1) / 2];
+    model->anchorObject->position.x = x + gx;
+    model->anchorObject->position.y = y;
 
     model->anchorObject->immobile = TRUE;
 }
@@ -957,29 +1054,25 @@ modelAddEdgeAnchors (Model *model,
 		     int   height)
 {
     Object *o;
-    float  x0, y0;
-
-    x0 = model->scaleOrigin.x;
-    y0 = model->scaleOrigin.y;
 
     o = &model->objects[0];
-    o->position.x = x + (0      - x0) * model->scale.x + x0;
-    o->position.y = y + (0      - y0) * model->scale.y + y0;
+    o->position.x = x;
+    o->position.y = y;
     o->immobile = TRUE;
 
     o = &model->objects[GRID_WIDTH - 1];
-    o->position.x = x + (width  - x0) * model->scale.x + x0;
-    o->position.y = y + (0      - y0) * model->scale.y + y0;
+    o->position.x = x + width;
+    o->position.y = y;
     o->immobile = TRUE;
 
     o = &model->objects[GRID_WIDTH * (GRID_HEIGHT - 1)];
-    o->position.x = x + (0      - x0) * model->scale.x + x0;
-    o->position.y = y + (height - y0) * model->scale.y + y0;
+    o->position.x = x;
+    o->position.y = y + height;
     o->immobile = TRUE;
 
     o = &model->objects[model->numObjects - 1];
-    o->position.x = x + (width  - x0) * model->scale.x + x0;
-    o->position.y = y + (height - y0) * model->scale.y + y0;
+    o->position.x = x + width;
+    o->position.y = y + height;
     o->immobile = TRUE;
 
     if (!model->anchorObject)
@@ -994,32 +1087,28 @@ modelRemoveEdgeAnchors (Model *model,
 			int   height)
 {
     Object *o;
-    float  x0, y0;
-
-    x0 = model->scaleOrigin.x;
-    y0 = model->scaleOrigin.y;
 
     o = &model->objects[0];
-    o->position.x = x + (0      - x0) * model->scale.x + x0;
-    o->position.y = y + (0      - y0) * model->scale.y + y0;
+    o->position.x = x;
+    o->position.y = y;
     if (o != model->anchorObject)
 	o->immobile = FALSE;
 
     o = &model->objects[GRID_WIDTH - 1];
-    o->position.x = x + (width  - x0) * model->scale.x + x0;
-    o->position.y = y + (0      - y0) * model->scale.y + y0;
+    o->position.x = x + width;
+    o->position.y = y;
     if (o != model->anchorObject)
 	o->immobile = FALSE;
 
     o = &model->objects[GRID_WIDTH * (GRID_HEIGHT - 1)];
-    o->position.x = x + (0      - x0) * model->scale.x + x0;
-    o->position.y = y + (height - y0) * model->scale.y + y0;
+    o->position.x = x;
+    o->position.y = y + height;
     if (o != model->anchorObject)
 	o->immobile = FALSE;
 
     o = &model->objects[model->numObjects - 1];
-    o->position.x = x + (width  - x0) * model->scale.x + x0;
-    o->position.y = y + (height - y0) * model->scale.y + y0;
+    o->position.x = x + width;
+    o->position.y = y + height;
     if (o != model->anchorObject)
 	o->immobile = FALSE;
 }
@@ -1033,12 +1122,8 @@ modelAdjustObjectPosition (Model  *model,
 			   int    height)
 {
     Object *o;
-    float  x0, y0;
     int	   gridX, gridY, i = 0;
 
-    x0 = model->scaleOrigin.x;
-    y0 = model->scaleOrigin.y;
-
     for (gridY = 0; gridY < GRID_HEIGHT; gridY++)
     {
 	for (gridX = 0; gridX < GRID_WIDTH; gridX++)
@@ -1046,12 +1131,8 @@ modelAdjustObjectPosition (Model  *model,
 	    o = &model->objects[i];
 	    if (o == object)
 	    {
-		o->position.x = x +
-		    (((gridX * width) / (GRID_WIDTH - 1)) - x0) *
-		    model->scale.x + x0;
-		o->position.y = y +
-		    (((gridY * height) / (GRID_HEIGHT - 1)) - y0) *
-		    model->scale.y + y0;
+		o->position.x = x + (gridX * width) / (GRID_WIDTH - 1);
+		o->position.y = y + (gridY * height) / (GRID_HEIGHT - 1);
 
 		return;
 	    }
@@ -1069,10 +1150,7 @@ modelInitObjects (Model *model,
 		  int	height)
 {
     int	  gridX, gridY, i = 0;
-    float gw, gh, x0, y0;
-
-    x0 = model->scaleOrigin.x;
-    y0 = model->scaleOrigin.y;
+    float gw, gh;
 
     gw = GRID_WIDTH  - 1;
     gh = GRID_HEIGHT - 1;
@@ -1082,10 +1160,8 @@ modelInitObjects (Model *model,
 	for (gridX = 0; gridX < GRID_WIDTH; gridX++)
 	{
 	    objectInit (&model->objects[i],
-			x + (((gridX * width)  / gw) - x0) *
-			model->scale.x + x0,
-			y + (((gridY * height) / gh) - y0) *
-			model->scale.y + y0,
+			x + (gridX * width) / gw,
+			y + (gridY * height) / gh,
 			0, 0);
 	    i++;
 	}
@@ -1225,11 +1301,11 @@ modelAdjustObjectsForShiver (Model *model,
 {
     int   gridX, gridY, i = 0;
     float vX, vY;
-    float scale;
     float w, h;
+    float scale;
 
-    w = (float) width  * model->scale.x;
-    h = (float) height * model->scale.y;
+    w = width;
+    h = height;
 
     for (gridY = 0; gridY < GRID_HEIGHT; gridY++)
     {
@@ -1266,8 +1342,8 @@ modelInitSprings (Model *model,
 
     model->numSprings = 0;
 
-    hpad = ((float) width  * model->scale.x) / (GRID_WIDTH  - 1);
-    vpad = ((float) height * model->scale.y) / (GRID_HEIGHT - 1);
+    hpad = ((float) width) / (GRID_WIDTH  - 1);
+    vpad = ((float) height) / (GRID_HEIGHT - 1);
 
     for (gridY = 0; gridY < GRID_HEIGHT; gridY++)
     {
@@ -1327,14 +1403,6 @@ createModel (int	  x,
 
     model->steps = 0;
 
-    model->scale.x = 1.0f;
-    model->scale.y = 1.0f;
-
-    model->scaleOrigin.x = 0.0f;
-    model->scaleOrigin.y = 0.0f;
-
-    model->transformed = FALSE;
-
     memset (model->snapCnt, 0, sizeof (model->snapCnt));
 
     model->edgeMask = edgeMask;
@@ -1837,7 +1905,7 @@ wobblyPreparePaintScreen (CompScreen *s,
 
     if (ws->wobblyWindows & (WobblyInitial | WobblyVelocity))
     {
-	REGION region;
+	BoxRec box;
 	Point  topLeft, bottomRight;
 	float  friction, springK;
 	Model  *model;
@@ -1845,9 +1913,6 @@ wobblyPreparePaintScreen (CompScreen *s,
 	friction = ws->opt[WOBBLY_SCREEN_OPTION_FRICTION].value.f;
 	springK  = ws->opt[WOBBLY_SCREEN_OPTION_SPRING_K].value.f;
 
-	region.rects = &region.extents;
-	region.numRects = region.size = 1;
-
 	ws->wobblyWindows = 0;
 	for (w = s->windows; w; w = w->next)
 	{
@@ -1880,20 +1945,23 @@ wobblyPreparePaintScreen (CompScreen *s,
 			    ww->wobbly |= WobblyInitial;
 			}
 		    }
-		    else if (!ww->model->transformed)
+		    else
 		    {
 			ww->model = 0;
 
-			moveWindow (w,
-				    model->topLeft.x + w->output.left -
-				    w->attrib.x,
-				    model->topLeft.y + w->output.top -
-				    w->attrib.y,
-				    TRUE, TRUE);
+			if (w->attrib.x == w->serverX &&
+			    w->attrib.y == w->serverY)
+			{
+			    moveWindow (w,
+					model->topLeft.x + w->output.left -
+					w->attrib.x,
+					model->topLeft.y + w->output.top -
+					w->attrib.y,
+					TRUE, TRUE);
+			    syncWindowPosition (w);
+			}
 
 			ww->model = model;
-
-			syncWindowPosition (w);
 		    }
 
 		    if (!(s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK))
@@ -1912,12 +1980,17 @@ wobblyPreparePaintScreen (CompScreen *s,
 			else
 			    addWindowDamage (w);
 
-			region.extents.x1 = topLeft.x;
-			region.extents.y1 = topLeft.y;
-			region.extents.x2 = bottomRight.x + 0.5f;
-			region.extents.y2 = bottomRight.y + 0.5f;
+			box.x1 = topLeft.x;
+			box.y1 = topLeft.y;
+			box.x2 = bottomRight.x + 0.5f;
+			box.y2 = bottomRight.y + 0.5f;
 
-			damageScreenRegion (s, &region);
+			box.x1 -= w->attrib.x + w->attrib.border_width;
+			box.y1 -= w->attrib.y + w->attrib.border_width;
+			box.x2 -= w->attrib.x + w->attrib.border_width;
+			box.y2 -= w->attrib.y + w->attrib.border_width;
+
+			addWindowDamageRect (w, &box);
 		    }
 		}
 
@@ -1944,6 +2017,46 @@ wobblyDonePaintScreen (CompScreen *s)
     WRAP (ws, s, donePaintScreen, wobblyDonePaintScreen);
 }
 
+static void
+wobblyDrawWindowGeometry (CompWindow *w)
+{
+    int     texUnit = w->texUnits;
+    int     currentTexUnit = 0;
+    int     stride = (1 + texUnit) * 2;
+    GLfloat *vertices = w->vertices + (stride - 2);
+
+    stride *= sizeof (GLfloat);
+
+    glVertexPointer (2, GL_FLOAT, stride, vertices);
+
+    while (texUnit--)
+    {
+	if (texUnit != currentTexUnit)
+	{
+	    w->screen->clientActiveTexture (GL_TEXTURE0_ARB + texUnit);
+	    glEnableClientState (GL_TEXTURE_COORD_ARRAY);
+	    currentTexUnit = texUnit;
+	}
+	vertices -= 2;
+	glTexCoordPointer (2, GL_FLOAT, stride, vertices);
+    }
+
+    glDrawElements (GL_QUADS, w->indexCount, GL_UNSIGNED_SHORT, w->indices);
+
+    /* disable all texture coordinate arrays except 0 */
+    texUnit = w->texUnits;
+    if (texUnit > 1)
+    {
+	while (--texUnit)
+	{
+	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
+	    glDisableClientState (GL_TEXTURE_COORD_ARRAY);
+	}
+
+	(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
+    }
+}
+
 static void
 wobblyAddWindowGeometry (CompWindow *w,
 			 CompMatrix *matrix,
@@ -1998,7 +2111,7 @@ wobblyAddWindowGeometry (CompWindow *w,
 	vSize = 2 + nMatrix * 2;
 
 	nVertices = w->vCount;
-	nIndices  = w->vCount;
+	nIndices  = w->indexCount;
 
 	v = w->vertices + (nVertices * vSize);
 	i = w->indices  + nIndices;
@@ -2094,7 +2207,9 @@ wobblyAddWindowGeometry (CompWindow *w,
 	    pClip++;
 	}
 
-	w->vCount = nIndices;
+	w->vCount	      = nVertices;
+	w->indexCount	      = nIndices;
+	w->drawWindowGeometry = wobblyDrawWindowGeometry;
     }
     else
     {
@@ -2104,91 +2219,10 @@ wobblyAddWindowGeometry (CompWindow *w,
     }
 }
 
-static void
-wobblyDrawWindowGeometry (CompWindow *w)
-{
-    WOBBLY_WINDOW (w);
-
-    if (ww->wobbly)
-    {
-	int     texUnit = w->texUnits;
-	int     currentTexUnit = 0;
-	int     stride = (1 + texUnit) * 2;
-	GLfloat *vertices = w->vertices + (stride - 2);
-
-	stride *= sizeof (GLfloat);
-
-	glVertexPointer (2, GL_FLOAT, stride, vertices);
-
-	while (texUnit--)
-	{
-	    if (texUnit != currentTexUnit)
-	    {
-		w->screen->clientActiveTexture (GL_TEXTURE0_ARB + texUnit);
-		glEnableClientState (GL_TEXTURE_COORD_ARRAY);
-		currentTexUnit = texUnit;
-	    }
-	    vertices -= 2;
-	    glTexCoordPointer (2, GL_FLOAT, stride, vertices);
-	}
-
-	glDrawElements (GL_QUADS, w->vCount, GL_UNSIGNED_SHORT, w->indices);
-
-	/* disable all texture coordinate arrays except 0 */
-	texUnit = w->texUnits;
-	if (texUnit > 1)
-	{
-	    while (--texUnit)
-	    {
-		(*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB + texUnit);
-		glDisableClientState (GL_TEXTURE_COORD_ARRAY);
-	    }
-
-	    (*w->screen->clientActiveTexture) (GL_TEXTURE0_ARB);
-	}
-    }
-    else
-    {
-	WOBBLY_SCREEN (w->screen);
-
-	UNWRAP (ws, w->screen, drawWindowGeometry);
-	(*w->screen->drawWindowGeometry) (w);
-	WRAP (ws, w->screen, drawWindowGeometry, wobblyDrawWindowGeometry);
-    }
-}
-
-static void
-wobblyDrawWindowTexture (CompWindow		 *w,
-			 CompTexture		 *texture,
-			 const WindowPaintAttrib *attrib,
-			 unsigned int		 mask)
-{
-    WOBBLY_SCREEN (w->screen);
-    WOBBLY_WINDOW (w);
-
-    if (ww->wobbly)
-    {
-	WindowPaintAttrib wAttrib = *attrib;
-
-	/* remove scale that is applied at the add geometry stage */
-	wAttrib.xScale = attrib->xScale / ww->model->scale.x;
-	wAttrib.yScale = attrib->yScale / ww->model->scale.y;
-
-	UNWRAP (ws, w->screen, drawWindowTexture);
-	(*w->screen->drawWindowTexture) (w, texture, &wAttrib, mask);
-	WRAP (ws, w->screen, drawWindowTexture, wobblyDrawWindowTexture);
-    }
-    else
-    {
-	UNWRAP (ws, w->screen, drawWindowTexture);
-	(*w->screen->drawWindowTexture) (w, texture, attrib, mask);
-	WRAP (ws, w->screen, drawWindowTexture, wobblyDrawWindowTexture);
-    }
-}
-
 static Bool
 wobblyPaintWindow (CompWindow		   *w,
 		   const WindowPaintAttrib *attrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   unsigned int		   mask)
 {
@@ -2199,14 +2233,14 @@ wobblyPaintWindow (CompWindow		   *w,
 
     if (ww->wobbly)
     {
-	if (mask & PAINT_WINDOW_SOLID_MASK)
+	if (mask & PAINT_WINDOW_CLIP_OPAQUE_MASK)
 	    return FALSE;
 
 	mask |= PAINT_WINDOW_TRANSFORMED_MASK;
     }
 
     UNWRAP (ws, w->screen, paintWindow);
-    status = (*w->screen->paintWindow) (w, attrib, region, mask);
+    status = (*w->screen->paintWindow) (w, attrib, transform, region, mask);
     WRAP (ws, w->screen, paintWindow, wobblyPaintWindow);
 
     return status;
@@ -2319,17 +2353,13 @@ static void
 wobblyHandleEvent (CompDisplay *d,
 		   XEvent      *event)
 {
-    Window     activeWindow = 0;
+    Window     activeWindow = d->activeWindow;
     CompWindow *w;
     CompScreen *s;
 
     WOBBLY_DISPLAY (d);
 
     switch (event->type) {
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	    activeWindow = d->activeWindow;
-	break;
     case MapNotify:
 	w = findWindowAtDisplay (d, event->xmap.window);
 	if (w)
@@ -2429,44 +2459,40 @@ wobblyHandleEvent (CompDisplay *d,
 		}
 	    }
 	}
-	break;
-    case PropertyNotify:
-	if (event->xproperty.atom == d->winActiveAtom)
-	{
-	    if (d->activeWindow != activeWindow)
-	    {
-		w = findWindowAtDisplay (d, d->activeWindow);
-		if (w && isWobblyWin (w))
-		{
-		    WOBBLY_WINDOW (w);
-		    WOBBLY_SCREEN (w->screen);
-
-		    if ((ws->focusWMask & w->type) &&
-			ws->focusEffect		   &&
-			wobblyEnsureModel (w))
-		    {
-			switch (ws->focusEffect) {
-			case WobblyEffectShiver:
-			    modelAdjustObjectsForShiver (ww->model,
-							 WIN_X (w),
-							 WIN_Y (w),
-							 WIN_W (w),
-							 WIN_H (w));
-			default:
-			    break;
-			}
-
-			ww->wobbly |= WobblyInitial;
-			ws->wobblyWindows |= ww->wobbly;
-
-			damagePendingOnScreen (w->screen);
-		    }
-		}
-	    }
-	}
     default:
 	break;
     }
+
+    if (d->activeWindow != activeWindow)
+    {
+	w = findWindowAtDisplay (d, d->activeWindow);
+	if (w && isWobblyWin (w))
+	{
+	    WOBBLY_WINDOW (w);
+	    WOBBLY_SCREEN (w->screen);
+
+	    if ((ws->focusWMask & w->type) &&
+		ws->focusEffect		   &&
+		wobblyEnsureModel (w))
+	    {
+		switch (ws->focusEffect) {
+		case WobblyEffectShiver:
+		    modelAdjustObjectsForShiver (ww->model,
+						 WIN_X (w),
+						 WIN_Y (w),
+						 WIN_W (w),
+						 WIN_H (w));
+		default:
+		    break;
+		}
+
+		ww->wobbly |= WobblyInitial;
+		ws->wobblyWindows |= ww->wobbly;
+
+		damagePendingOnScreen (w->screen);
+	    }
+	}
+    }
 }
 
 static Bool
@@ -2538,45 +2564,6 @@ wobblyDamageWindowRect (CompWindow *w,
     return status;
 }
 
-static void
-wobblySetWindowScale (CompWindow *w,
-		      float      xScale,
-		      float      yScale)
-{
-    WOBBLY_WINDOW (w);
-    WOBBLY_SCREEN (w->screen);
-
-    UNWRAP (ws, w->screen, setWindowScale);
-    (*w->screen->setWindowScale) (w, xScale, yScale);
-    WRAP (ws, w->screen, setWindowScale, wobblySetWindowScale);
-
-    if (wobblyEnsureModel (w))
-    {
-	if (ww->model->scale.x != xScale ||
-	    ww->model->scale.y != yScale)
-	{
-	    ww->model->scale.x = xScale;
-	    ww->model->scale.y = yScale;
-
-	    ww->model->scaleOrigin.x = w->output.left;
-	    ww->model->scaleOrigin.y = w->output.top;
-
-	    modelInitObjects (ww->model,
-			      WIN_X (w), WIN_Y (w),
-			      WIN_W (w), WIN_H (w));
-
-	    modelInitSprings (ww->model,
-			      WIN_X (w), WIN_Y (w),
-			      WIN_W (w), WIN_H (w));
-	}
-
-	if (ww->model->scale.x != 1.0f || ww->model->scale.y != 1.0f)
-	    ww->model->transformed = 1;
-	else
-	    ww->model->transformed = 0;
-    }
-}
-
 static void
 wobblyWindowResizeNotify (CompWindow *w)
 {
@@ -2585,8 +2572,11 @@ wobblyWindowResizeNotify (CompWindow *w)
 
     if (ws->opt[WOBBLY_SCREEN_OPTION_MAXIMIZE_EFFECT].value.b &&
 	isWobblyWin (w)					      &&
-	((w->state ^ ww->state) & MAXIMIZE_STATE))
+	((w->state | ww->state) & MAXIMIZE_STATE))
     {
+	ww->state &= ~MAXIMIZE_STATE;
+	ww->state |= w->state & MAXIMIZE_STATE;
+
 	if (wobblyEnsureModel (w))
 	{
 	    if (w->state & MAXIMIZE_STATE)
@@ -2622,8 +2612,16 @@ wobblyWindowResizeNotify (CompWindow *w)
     }
     else if (ww->model)
     {
-	modelInitObjects (ww->model,
-			  WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
+	if (ww->wobbly)
+	{
+	    if (!(ww->state & MAXIMIZE_STATE))
+		modelSetTopAnchor (ww->model, WIN_X (w), WIN_Y (w), WIN_W (w));
+	}
+	else
+	{
+	    modelInitObjects (ww->model,
+			      WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
+	}
 
 	modelInitSprings (ww->model,
 			  WIN_X (w), WIN_Y (w), WIN_W (w), WIN_H (w));
@@ -2646,8 +2644,6 @@ wobblyWindowResizeNotify (CompWindow *w)
 				   WIN_W (w), WIN_H (w));
     }
 
-    ww->state = w->state;
-
     UNWRAP (ws, w->screen, windowResizeNotify);
     (*w->screen->windowResizeNotify) (w);
     WRAP (ws, w->screen, windowResizeNotify, wobblyWindowResizeNotify);
@@ -2835,6 +2831,7 @@ wobblyWindowUngrabNotify (CompWindow *w)
 static Bool
 wobblyPaintScreen (CompScreen		   *s,
 		   const ScreenPaintAttrib *sAttrib,
+		   const CompTransform	   *transform,
 		   Region		   region,
 		   int			   output,
 		   unsigned int		   mask)
@@ -2847,7 +2844,7 @@ wobblyPaintScreen (CompScreen		   *s,
 	mask |= PAINT_SCREEN_WITH_TRANSFORMED_WINDOWS_MASK;
 
     UNWRAP (ws, s, paintScreen);
-    status = (*s->paintScreen) (s, sAttrib, region, output, mask);
+    status = (*s->paintScreen) (s, sAttrib, transform, region, output, mask);
     WRAP (ws, s, paintScreen, wobblyPaintScreen);
 
     return status;
@@ -3016,9 +3013,6 @@ wobblyInitScreen (CompPlugin *p,
     WRAP (ws, s, paintWindow, wobblyPaintWindow);
     WRAP (ws, s, damageWindowRect, wobblyDamageWindowRect);
     WRAP (ws, s, addWindowGeometry, wobblyAddWindowGeometry);
-    WRAP (ws, s, drawWindowGeometry, wobblyDrawWindowGeometry);
-    WRAP (ws, s, drawWindowTexture, wobblyDrawWindowTexture);
-    WRAP (ws, s, setWindowScale, wobblySetWindowScale);
     WRAP (ws, s, windowResizeNotify, wobblyWindowResizeNotify);
     WRAP (ws, s, windowMoveNotify, wobblyWindowMoveNotify);
     WRAP (ws, s, windowGrabNotify, wobblyWindowGrabNotify);
@@ -3046,9 +3040,6 @@ wobblyFiniScreen (CompPlugin *p,
     UNWRAP (ws, s, paintWindow);
     UNWRAP (ws, s, damageWindowRect);
     UNWRAP (ws, s, addWindowGeometry);
-    UNWRAP (ws, s, drawWindowGeometry);
-    UNWRAP (ws, s, drawWindowTexture);
-    UNWRAP (ws, s, setWindowScale);
     UNWRAP (ws, s, windowResizeNotify);
     UNWRAP (ws, s, windowMoveNotify);
     UNWRAP (ws, s, windowGrabNotify);
