@@ -86,6 +86,15 @@ static Bool inHandleEvent = FALSE;
 
 static Bool shutDown = FALSE;
 
+static const CompTransform identity = {
+    {
+	1.0, 0.0, 0.0, 0.0,
+	0.0, 1.0, 0.0, 0.0,
+	0.0, 0.0, 1.0, 0.0,
+	0.0, 0.0, 0.0, 1.0
+    }
+};
+
 int lastPointerX = 0;
 int lastPointerY = 0;
 int pointerX     = 0;
@@ -158,12 +167,16 @@ int pointerY     = 0;
 
 #define IGNORE_HINTS_WHEN_MAXIMIZED_DEFAULT TRUE
 
-#define NUM_OPTIONS(d) (sizeof ((d)->opt) / sizeof (CompOption))
+#define PING_DELAY_DEFAULT 5000
+#define PING_DELAY_MIN	   1000
+#define PING_DELAY_MAX	   30000
 
 static char *textureFilter[] = { N_("Fast"), N_("Good"), N_("Best") };
 
 #define NUM_TEXTURE_FILTER (sizeof (textureFilter) / sizeof (textureFilter[0]))
 
+#define NUM_OPTIONS(d) (sizeof ((d)->opt) / sizeof (CompOption))
+
 CompDisplay *compDisplays = 0;
 
 static CompDisplay compDisplay;
@@ -594,6 +607,25 @@ runCommandWindowScreenshot (CompDisplay     *d,
     return TRUE;
 }
 
+static Bool
+runCommandTerminal (CompDisplay     *d,
+		    CompAction      *action,
+		    CompActionState state,
+		    CompOption      *option,
+		    int	            nOption)
+{
+    CompScreen *s;
+    Window     xid;
+
+    xid = getIntOptionNamed (option, nOption, "root", 0);
+
+    s = findScreenAtDisplay (d, xid);
+    if (s)
+	runCommand (s, d->opt[COMP_DISPLAY_OPTION_TERMINAL].value.s);
+
+    return TRUE;
+}
+
 static Bool
 windowMenu (CompDisplay     *d,
 	    CompAction      *action,
@@ -1197,6 +1229,37 @@ compDisplayInitOptions (CompDisplay *display,
 	"maximized");
     o->type	 = CompOptionTypeBool;
     o->value.b	 = IGNORE_HINTS_WHEN_MAXIMIZED_DEFAULT;
+
+    o = &display->opt[COMP_DISPLAY_OPTION_TERMINAL];
+    o->name	      = "command_terminal";
+    o->shortDesc      = N_("Terminal command line");
+    o->longDesc	      = N_("Terminal command line");
+    o->type	      = CompOptionTypeString;
+    o->value.s	      = strdup ("");
+    o->rest.s.string  = NULL;
+    o->rest.s.nString = 0;
+
+    o = &display->opt[COMP_DISPLAY_OPTION_RUN_TERMINAL];
+    o->name		      = "run_command_terminal";
+    o->shortDesc	      = N_("Open a terminal");
+    o->longDesc		      = N_("Open a terminal");
+    o->type		      = CompOptionTypeAction;
+    o->value.action.initiate  = runCommandTerminal;
+    o->value.action.terminate = 0;
+    o->value.action.bell      = FALSE;
+    o->value.action.edgeMask  = 0;
+    o->value.action.state     = CompActionStateInitKey;
+    o->value.action.state    |= CompActionStateInitButton;
+    o->value.action.type      = CompBindingTypeNone;
+
+    o = &display->opt[COMP_DISPLAY_OPTION_PING_DELAY];
+    o->name	  = "ping_delay";
+    o->shortDesc  = N_("Ping Delay");
+    o->longDesc	  = N_("Interval between ping messages");
+    o->type	  = CompOptionTypeInt;
+    o->value.i	  = PING_DELAY_DEFAULT;
+    o->rest.i.min = PING_DELAY_MIN;
+    o->rest.i.max = PING_DELAY_MAX;
 }
 
 CompOption *
@@ -1218,6 +1281,78 @@ setAudibleBell (CompDisplay *display,
 				  audible ? XkbAudibleBellMask : 0);
 }
 
+static Bool
+pingTimeout (void *closure)
+{
+    CompDisplay *d = closure;
+    CompScreen  *s;
+    CompWindow  *w;
+    XEvent      ev;
+    int		ping = d->lastPing + 1;
+
+    ev.type		    = ClientMessage;
+    ev.xclient.window	    = 0;
+    ev.xclient.message_type = d->wmProtocolsAtom;
+    ev.xclient.format	    = 32;
+    ev.xclient.data.l[0]    = d->wmPingAtom;
+    ev.xclient.data.l[1]    = ping;
+    ev.xclient.data.l[2]    = 0;
+    ev.xclient.data.l[3]    = 0;
+    ev.xclient.data.l[4]    = 0;
+
+    for (s = d->screens; s; s = s->next)
+    {
+	for (w = s->windows; w; w = w->next)
+	{
+	    if (w->attrib.map_state != IsViewable)
+		continue;
+
+	    if (!(w->type & CompWindowTypeNormalMask))
+		continue;
+
+	    if (w->protocols & CompWindowProtocolPingMask)
+	    {
+		if (w->transientFor)
+		    continue;
+
+		if (w->lastPong < d->lastPing)
+		{
+		    if (w->alive)
+		    {
+			w->alive	    = FALSE;
+			w->paint.brightness = 0xa8a8;
+			w->paint.saturation = 0;
+
+			if (w->closeRequests)
+			{
+			    toolkitAction (s,
+					   d->toolkitActionForceQuitDialogAtom,
+					   w->lastCloseRequestTime,
+					   w->id,
+					   TRUE,
+					   0,
+					   0);
+
+			    w->closeRequests = 0;
+			}
+
+			addWindowDamage (w);
+		    }
+		}
+
+		ev.xclient.window    = w->id;
+		ev.xclient.data.l[2] = w->id;
+
+		XSendEvent (d->display, w->id, FALSE, NoEventMask, &ev);
+	    }
+	}
+    }
+
+    d->lastPing = ping;
+
+    return TRUE;
+}
+
 static Bool
 setDisplayOption (CompDisplay     *display,
 		  char	          *name,
@@ -1266,6 +1401,17 @@ setDisplayOption (CompDisplay     *display,
 	if (compSetIntOption (o, value))
 	    return TRUE;
 	break;
+    case COMP_DISPLAY_OPTION_PING_DELAY:
+	if (compSetIntOption (o, value))
+	{
+	    if (display->pingHandle)
+		compRemoveTimeout (display->pingHandle);
+
+	    display->pingHandle =
+		compAddTimeout (o->value.i, pingTimeout, display);
+	    return TRUE;
+	}
+	break;
     case COMP_DISPLAY_OPTION_COMMAND0:
     case COMP_DISPLAY_OPTION_COMMAND1:
     case COMP_DISPLAY_OPTION_COMMAND2:
@@ -1280,6 +1426,7 @@ setDisplayOption (CompDisplay     *display,
     case COMP_DISPLAY_OPTION_COMMAND11:
     case COMP_DISPLAY_OPTION_SCREENSHOT:
     case COMP_DISPLAY_OPTION_WINDOW_SCREENSHOT:
+    case COMP_DISPLAY_OPTION_TERMINAL:
 	if (compSetStringOption (o, value))
 	    return TRUE;
 	break;
@@ -1311,6 +1458,7 @@ setDisplayOption (CompDisplay     *display,
     case COMP_DISPLAY_OPTION_OPACITY_DECREASE:
     case COMP_DISPLAY_OPTION_RUN_SCREENSHOT:
     case COMP_DISPLAY_OPTION_RUN_WINDOW_SCREENSHOT:
+    case COMP_DISPLAY_OPTION_RUN_TERMINAL:
     case COMP_DISPLAY_OPTION_WINDOW_MENU:
     case COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED:
     case COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ:
@@ -2018,8 +2166,6 @@ eventLoop (void)
 		    if (timeDiff < 0)
 			timeDiff = 0;
 
-		    s->stencilRef = 0;
-
 		    makeScreenCurrent (s);
 
 		    if (s->slowAnimations)
@@ -2094,6 +2240,7 @@ eventLoop (void)
 			{
 			    (*s->paintScreen) (s,
 					       &defaultScreenPaintAttrib,
+					       &identity,
 					       &s->outputDev[i].region, i,
 					       PAINT_SCREEN_REGION_MASK |
 					       PAINT_SCREEN_FULL_MASK);
@@ -2106,11 +2253,13 @@ eventLoop (void)
 
 			    if (!(*s->paintScreen) (s,
 						    &defaultScreenPaintAttrib,
+						    &identity,
 						    outputRegion, i,
 						    PAINT_SCREEN_REGION_MASK))
 			    {
 				(*s->paintScreen) (s,
 						   &defaultScreenPaintAttrib,
+						   &identity,
 						   &s->outputDev[i].region, i,
 						   PAINT_SCREEN_FULL_MASK);
 
@@ -2310,80 +2459,6 @@ compCheckForError (Display *dpy)
     return e;
 }
 
-#define PING_DELAY 5000
-
-static Bool
-pingTimeout (void *closure)
-{
-    CompDisplay *d = closure;
-    CompScreen  *s;
-    CompWindow  *w;
-    XEvent      ev;
-    int		ping = d->lastPing + 1;
-
-    ev.type		    = ClientMessage;
-    ev.xclient.window	    = 0;
-    ev.xclient.message_type = d->wmProtocolsAtom;
-    ev.xclient.format	    = 32;
-    ev.xclient.data.l[0]    = d->wmPingAtom;
-    ev.xclient.data.l[1]    = ping;
-    ev.xclient.data.l[2]    = 0;
-    ev.xclient.data.l[3]    = 0;
-    ev.xclient.data.l[4]    = 0;
-
-    for (s = d->screens; s; s = s->next)
-    {
-	for (w = s->windows; w; w = w->next)
-	{
-	    if (w->attrib.map_state != IsViewable)
-		continue;
-
-	    if (!(w->type & CompWindowTypeNormalMask))
-		continue;
-
-	    if (w->protocols & CompWindowProtocolPingMask)
-	    {
-		if (w->transientFor)
-		    continue;
-
-		if (w->lastPong < d->lastPing)
-		{
-		    if (w->alive)
-		    {
-			w->alive	    = FALSE;
-			w->paint.brightness = 0xa8a8;
-			w->paint.saturation = 0;
-
-			if (w->closeRequests)
-			{
-			    toolkitAction (s,
-					   d->toolkitActionForceQuitDialogAtom,
-					   w->lastCloseRequestTime,
-					   w->id,
-					   TRUE,
-					   0,
-					   0);
-
-			    w->closeRequests = 0;
-			}
-
-			addWindowDamage (w);
-		    }
-		}
-
-		ev.xclient.window    = w->id;
-		ev.xclient.data.l[2] = w->id;
-
-		XSendEvent (d->display, w->id, FALSE, NoEventMask, &ev);
-	    }
-	}
-    }
-
-    d->lastPing = ping;
-
-    return TRUE;
-}
-
 static void
 addScreenActions (CompDisplay *d, CompScreen *s)
 {
@@ -2423,6 +2498,7 @@ addScreenActions (CompDisplay *d, CompScreen *s)
 		     &d->opt[COMP_DISPLAY_OPTION_OPACITY_DECREASE].value.action);
     addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_RUN_SCREENSHOT].value.action);
     addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_RUN_WINDOW_SCREENSHOT].value.action);
+    addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_RUN_TERMINAL].value.action);
     addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_WINDOW_MENU].value.action);
     addScreenAction (s,
 		     &d->opt[COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED].value.action);
@@ -2993,7 +3069,9 @@ addDisplay (char *name,
 	    focusDefaultWindow (d);
     }
 
-    d->pingHandle = compAddTimeout (PING_DELAY, pingTimeout, d);
+    d->pingHandle =
+	compAddTimeout (d->opt[COMP_DISPLAY_OPTION_PING_DELAY].value.i,
+			pingTimeout, d);
 
     return TRUE;
 }
